"""
This example shows the trade-off (pareto frontier) of deficit against cost by altering a reservoir control curve.

Two types of control curve are possible. The first is a monthly control curve containing one value for each
month. The second is a harmonic control curve with cosine terms around a mean. Both Parameter objects
are part of pywr.parameters.

The example demonstrates the use of two different optimisation libraries: pygmo and platypus. The
choice of library can be made with a command line argument.

"""
import json
import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


def get_model_data(harmonic=True):

    with open('two_reservoir.json') as fh:
        data = json.load(fh)

    if harmonic:
        # Patch the control curve parameter
        data['parameters']['control_curve'] = {
            'type': 'AnnualHarmonicSeries',
            'mean': 0.5,
            'amplitudes': [0.5, 0.0],
            'phases': [0.0, 0.0],
            'mean_upper_bounds': 1.0,
            'amplitude_upper_bounds': 1.0,
            'is_variable': True
        }

    return data


def platypus_main(harmonic=False):
    import platypus
    from pywr.optimisation.platypus import PlatypusWrapper

    wrapper = PlatypusWrapper(get_model_data(harmonic=harmonic))

    with platypus.ProcessPoolEvaluator() as evaluator:
        algorithm = platypus.NSGAIII(wrapper.problem, population_size=50, evaluator=evaluator, divisions_outer=12)
        algorithm.run(10000)

    objectives = pd.DataFrame(data=np.array([s.objectives for s in algorithm.result]),
                              columns=[o.name for o in wrapper.model_objectives])

    title = 'harmonic' if harmonic else 'monthly'
    objectives.to_hdf('two_reservoir_moea.h5', f'platypus_{title}')


def pygmo_main(harmonic=False):
    import pygmo as pg
    from pywr.optimisation.pygmo import PygmoWrapper

    def update_archive(pop, archive=None):
        if archive is None:
            combined_f = pop.get_f()
        else:
            combined_f = np.r_[archive, pop.get_f()]
        indices = pg.select_best_N_mo(combined_f, 50)
        new_archive = combined_f[indices, :]
        new_archive = np.unique(new_archive.round(4), axis=0)
        return new_archive

    wrapper = PygmoWrapper(get_model_data(harmonic=harmonic))
    prob = pg.problem(wrapper)
    print(prob)
    algo = pg.algorithm(pg.moead(gen=1))
    #algo = pg.algorithm(pg.nsga2(gen=1))

    pg.mp_island.init_pool(2)
    isl = pg.island(algo=algo, prob=prob, size=50, udi=pg.mp_island())

    ref_point = [216500, 4000]
    pop = isl.get_population()

    print('Evolving!')
    archive = update_archive(pop)
    hv = pg.hypervolume(archive)
    vol = hv.compute(ref_point)
    hvs = [vol, ]
    print('Gen: {:03d}, Hypervolume: {:6.4g}, Archive size: {:d}'.format(0, vol, len(archive)))

    for gen in range(20):
        isl.evolve(1)
        isl.wait_check()

        pop = isl.get_population()
        archive = update_archive(pop, archive)

        hv = pg.hypervolume(archive)
        vol = hv.compute(ref_point)
        print('Gen: {:03d}, Hypervolume: {:6.4g}, Archive size: {:d}'.format(gen+1, vol, len(archive)))
        hvs.append(vol)

    hvs = pd.Series(hvs)

    print('Finished!')

    plt.scatter(archive[:, 0], archive[:, 1])

    objectives = wrapper.model_objectives
    plt.xlabel(objectives[0].name)
    plt.ylabel(objectives[1].name)
    plt.grid(True)
    title = 'Harmonic Control Curve' if harmonic else 'Monthly Control Curve'
    plt.savefig('{} Example ({}).pdf'.format('pygmo', title), format='pdf')

    fig, ax = plt.subplots()
    ax.plot(hvs/1e6, marker='o')
    ax.grid(True)
    ax.set_ylabel('Hypervolume')
    ax.set_xlabel('Generation')
    plt.savefig('{} Example Hypervolume ({}).pdf'.format('pygmo', title), format='pdf')

    plt.show()


if __name__ == '__main__':
    import argparse

    # Setup logging
    import logging
    logger = logging.getLogger('pywr')
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    parser = argparse.ArgumentParser()
    parser.add_argument('--harmonic', action='store_true', help='Use an harmonic control curve.')
    parser.add_argument('--library', type=str, choices=['platypus', 'pygmo'])
    args = parser.parse_args()

    if args.library == 'platypus':
        platypus_main(harmonic=args.harmonic)
    elif args.library == 'pygmo':
        pygmo_main(harmonic=args.harmonic)
    else:
        raise ValueError('Optimisation library "{}" not recognised.'.format(args.library))



import cython
import numpy as np
cimport numpy as np
from .parameters import parameter_registry, ConstantParameter, parameter_property
from ._parameters import load_parameter, load_parameter_values, Parameter, IndexParameter
import warnings

# http://stackoverflow.com/a/20031818/1300519
cdef extern from "numpy/npy_math.h":
    bint npy_isnan(double x)


@cython.cdivision(True)
cpdef double _interpolate(double current_position, double lower_bound, double upper_bound, double lower_value, double upper_value):
    """Interpolation function used by PiecewiseLinearControlCurve"""
    cdef double factor
    cdef double value
    if current_position < lower_bound:
        value = lower_value
    elif current_position > upper_bound:
        value = upper_value
    elif upper_bound == lower_bound:
        value = lower_value
    else:
        factor = (current_position - lower_bound) / (upper_bound - lower_bound)
        value = lower_value + (upper_value - lower_value) * factor
    return value


cdef class BaseControlCurveParameter(Parameter):
    """ Base class for all Parameters that rely on a the attached Node containing a control_curve Parameter

    """
    def __init__(self, model, AbstractStorage storage_node, control_curves, **kwargs):
        """

        Parameters
        ----------
        storage_node : `Storage`
            An optional `Storage` node that can be used to query the current percentage volume.
        control_curves : iterable of Parameter objects or single Parameter
            The Parameter objects to use as a control curve(s).
        """
        super(BaseControlCurveParameter, self).__init__(model, **kwargs)
        self.control_curves = control_curves
        if storage_node is None:
            raise ValueError("storage_node is required")
        self._storage_node = storage_node

    property control_curves:
        def __get__(self):
            return self._control_curves
        def __set__(self, control_curves):
            # Accept a single Parameter and convert to a list internally
            if isinstance(control_curves, Parameter):
                control_curves = [control_curves]

            # remove existing control curves (if any)
            if self._control_curves is not None:
                for control_curve in self._control_curves:
                    control_curve.parents.remove(self)

            _new_control_curves = []
            for control_curve in control_curves:
                # Accept numeric inputs and convert to `ConstantParameter`
                if isinstance(control_curve, (float, int)):
                    control_curve = ConstantParameter(self.model, control_curve)

                control_curve.parents.add(self)
                _new_control_curves.append(control_curve)
            self._control_curves = list(_new_control_curves)

    property storage_node:
        def __get__(self):
            return self._storage_node
        def __set__(self, value):
            self._storage_node = value

    @classmethod
    def _load_control_curves(cls, model, data):
        """ Private class method to load control curve data from dict. """

        control_curves = []
        if 'control_curve' in data:
            control_curves.append(load_parameter(model, data.pop('control_curve')))
        elif 'control_curves' in data:
            for pdata in data.pop('control_curves'):
                control_curves.append(load_parameter(model, pdata))
        return control_curves

    @classmethod
    def _load_storage_node(cls, model, data):
        """ Private class method to load storage node from dict. """
        node = model.nodes[data.pop("storage_node")]
        return node


BaseControlCurveParameter.register()


cdef class ControlCurveInterpolatedParameter(BaseControlCurveParameter):
    """A control curve Parameter that interpolates between three or more values

    Return values are linearly interpolated between control curves, with the
    first and last value being 100% and 0% respectively.

    Parameters
    ----------
    storage_node : `Storage`
        The storage node to compare the control curve(s) to.
    control_curves : list of `Parameter` or floats
        A list of parameters representing the control curve(s). These are
        often MonthlyProfileParameters or DailyProfileParameters, but may be
        any Parameter that returns values between 0.0 and 1.0. If floats are
        passed they are converted to `ConstantParameter`.
    values : list of float
        A list of values to return corresponding to the control curves. The
        length of the list should be 2 + len(control_curves).
    parameters : iterable `Parameter` objects or `None`, optional
        If `values` is `None` then `parameters` can specify a `Parameter` object to use at each
        of the control curves. The number of parameters should be 2 + len(control_curves)

    Examples
    --------
    In the example below the cost of a storage node is related to it's volume.
    At 100% full the cost is 0. Between 100% and 50% the cost is linearly
    interpolated between 0 and -5. Between 50% and 30% the cost is interpolated
    between -5 and -10. Between 30% and 0% the cost is interpolated between -10
    and -20

    ::

        Volume:  100%             50%      30%       0%
                  |----------------|--------|--------|
          Cost:  0.0            -5.0     -10.0   -20.0


    >>> storage_node = Storage(model, "reservoir", max_volume=100, initial_volume=100)
    >>> ccs = [ConstantParameter(0.5), ConstantParameter(0.3)]
    >>> values = [0.0, -5.0, -10.0, -20.0]
    >>> cost = ControlCurveInterpolatedParameter(storage_node, ccs, values)
    >>> storage_node.cost = cost
    """
    def __init__(self, model, storage_node, control_curves, values=None, parameters=None, **kwargs):
        super(ControlCurveInterpolatedParameter, self).__init__(model, storage_node, control_curves, **kwargs)
        # Expected number of values is number of control curves plus two.
        nvalues = len(self.control_curves) + 2
        self.parameters = None

        if values is not None:
            if len(values) != nvalues:
                raise ValueError('Length of values should be two more than the number of '
                                 'control curves ({}).'.format(nvalues))
            self.values = np.asarray(values)

        elif parameters is not None:
            if len(parameters) != nvalues:
                raise ValueError('Length of parameters should be two more than the number of '
                                 'control curves ({}).'.format(nvalues))
            self.parameters = list(parameters)
            # Make sure these parameters depend on this parameter to ensure they are evaluated
            # in the correct order.
            for p in self.parameters:
                p.parents.add(self)
        else:
            raise ValueError('One of values or parameters keywords must be given.')

    property values:
        def __get__(self):
            return np.array(self._values)
        def __set__(self, values):
            self._values = np.array(values)

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        cdef int j
        cdef Parameter cc_param, value_param
        cdef double cc, cc_prev
        # return the interpolated value for the current level.
        cdef double current_pc = self._storage_node._current_pc[scenario_index.global_id]
        cdef double weight
        cdef double[:] values  # y values to interpolate between in this time-step

        if self.parameters is not None:
            # If there are parameter use them to gather the interpolation values
            values = np.empty(len(self.parameters))
            for j, value_param in enumerate(self.parameters):
                values[j] = value_param.get_value(scenario_index)
        else:
            # Otherwise use the given array of floats.
            # This makes a reference rather than a copy.
            values = self._values

        # Bounds check the current pc storage.
        # Always return the first value if storage above 100% or NaN
        if current_pc > 1.0 or npy_isnan(current_pc):
            return values[0]
        # Always return last value is storage less than 0%
        if current_pc < 0.0:
            return values[-1]

        # Assumes control_curves is sorted highest to lowest
        # First level 100%
        cc_prev = 1.0
        for j, cc_param in enumerate(self._control_curves):
            cc = cc_param.get_value(scenario_index)
            # If level above control curve then return this level's value
            if current_pc >= cc:
                try:
                    weight = (current_pc - cc) / (cc_prev - cc)
                except ZeroDivisionError:
                    # Last two control curves identical; return the next value
                    return values[j+1]
                return values[j]*weight + values[j+1]*(1.0 - weight)
            # Update previous value for next iteration
            cc_prev = cc

        # Current storage is above none of the control curves
        # Therefore interpolate between last control curve and bottom
        cc = 0.0
        try:
            weight = (current_pc - cc) / (cc_prev - cc)
        except ZeroDivisionError:
            # cc_prev == cc  i.e. last control curve is close to 0%
            return values[-2]
        return values[-2]*weight + values[-1]*(1.0 - weight)

    @classmethod
    def load(cls, model, data):
        control_curves = super(ControlCurveInterpolatedParameter, cls)._load_control_curves(model, data)
        storage_node = super(ControlCurveInterpolatedParameter, cls)._load_storage_node(model, data)
        if "parameters" in data:
            parameters = [load_parameter(model, p) for p in data.pop("parameters")]
            values = None
        else:
            values = load_parameter_values(model, data)
            parameters = None
        return cls(model, storage_node, control_curves, values=values, parameters=parameters)

ControlCurveInterpolatedParameter.register()


cdef class ControlCurvePiecewiseInterpolatedParameter(BaseControlCurveParameter):
    """A control curve Parameter that interpolates between two or more pairs of values.

    Return values are linearly interpolated between a pair of values depending on the current
    storage. The first pair is used between maximum and the first control curve, the next pair
    between the first control curve and second control curve, and so on until the last pair is
    used between the last control curve and the minimum value. The first value in each pair is the
    value at the upper position, and the second the value at the lower position.

    Parameters
    ----------
    storage_node : `Storage`
        The storage node to compare the control curve(s) to.
    control_curves : list of `Parameter` or floats
        A list of parameters representing the control curve(s). These are
        often MonthlyProfileParameters or DailyProfileParameters, but may be
        any Parameter that returns values between 0.0 and 1.0. If floats are
        passed they are converted to `ConstantParameter`.
    values : 2D array or list of lists
        A list of value pairs to interpolate between. The length of the list should be 1 + len(control_curves).
    minimum : float
        The storage considered the bottom of the lower curve, 0-1 (default=0).
    maximum : float
        The storage considered the top of the upper curve, 0-1 (default=1).

    """
    def __init__(self, model, storage_node, control_curves, values, minimum=0.0, maximum=1.0, **kwargs):
        super(ControlCurvePiecewiseInterpolatedParameter, self).__init__(model, storage_node, control_curves, **kwargs)
        self.values = np.array(values, dtype=np.float64)
        self.minimum = minimum
        self.maximum = maximum

    property values:
        def __get__(self):
            return np.array(self._values)
        def __set__(self, values):
            # Expected number of values is number of control curves plus one.
            nvalues = len(self.control_curves) + 1
            if len(values) != nvalues:
                raise ValueError('Length of values should be two more than the number of '
                                 'control curves ({}).'.format(nvalues))
            elif len(values[0]) != 2:
                raise ValueError('The second dimension of values should be of length 2.')
            self._values = np.array(values)

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        cdef int j, ncc
        cdef Parameter cc_param
        cdef double cc, cc_prev, val
        # return the interpolated value for the current level.
        cdef double current_pc = self._storage_node._current_pc[scenario_index.global_id]

        cc_prev = self.maximum
        for j, cc_param in enumerate(self._control_curves):
            cc = cc_param.get_value(scenario_index)
            # If level above control curve then return this level's value
            if current_pc >= cc:
                return _interpolate(current_pc, cc, cc_prev, self._values[j, 1], self._values[j, 0])

            # Update previous value for next iteration
            cc_prev = cc

        # Current storage is above none of the control curves
        # Therefore interpolate between last control curve and minimum
        ncc = len(self._control_curves)
        val = _interpolate(current_pc, self.minimum, cc_prev, self._values[ncc, 1], self._values[ncc, 0])
        return val

    @classmethod
    def load(cls, model, data):
        control_curves = super(ControlCurvePiecewiseInterpolatedParameter, cls)._load_control_curves(model, data)
        storage_node = super(ControlCurvePiecewiseInterpolatedParameter, cls)._load_storage_node(model, data)
        values = load_parameter_values(model, data)
        return cls(model, storage_node, control_curves, values=values, **data)

ControlCurvePiecewiseInterpolatedParameter.register()


cdef class ControlCurveIndexParameter(IndexParameter):
    """Multiple control curve holder which returns an index not a value

    Parameters
    ----------
    storage_node : `Storage`
    control_curves : iterable of `Parameter` instances or floats
    """
    def __init__(self, model, storage_node, control_curves, **kwargs):
        super(ControlCurveIndexParameter, self).__init__(model, **kwargs)
        self.storage_node = storage_node
        self.control_curves = control_curves

    property control_curves:
        def __get__(self):
            return self._control_curves
        def __set__(self, control_curves):
            # Accept a single Parameter and convert to a list internally
            if isinstance(control_curves, Parameter):
                control_curves = [control_curves]

            # remove existing control curves (if any)
            if self._control_curves is not None:
                for control_curve in self._control_curves:
                    control_curve.parents.remove(self)

            _new_control_curves = []
            for control_curve in control_curves:
                # Accept numeric inputs and convert to `ConstantParameter`
                if isinstance(control_curve, (float, int)):
                    control_curve = ConstantParameter(self.model, control_curve)

                control_curve.parents.add(self)
                _new_control_curves.append(control_curve)
            self._control_curves = list(_new_control_curves)

    cpdef int index(self, Timestep timestep, ScenarioIndex scenario_index) except? -1:
        """Returns the index of the first control curve the storage is above

        The index is zero-based. For example, if only one control curve is
        supplied then the index is either 0 (above) or 1 (below). For two
        curves the index is either 0 (above both), 1 (in between), or 2 (below
        both), and so on.
        """
        cdef double current_percentage
        cdef double target_percentage
        cdef int index, j
        cdef Parameter control_curve
        current_percentage = self.storage_node._current_pc[scenario_index.global_id]
        index = len(self.control_curves)
        for j, control_curve in enumerate(self.control_curves):
            target_percentage = control_curve.get_value(scenario_index)
            if current_percentage >= target_percentage:
                index = j
                break
        return index

    @classmethod
    def load(cls, model, data):
        storage_node = model.nodes[data.pop("storage_node")]
        control_curves = [load_parameter(model, d) for d in data.pop("control_curves")]
        return cls(model, storage_node, control_curves, **data)
ControlCurveIndexParameter.register()


cdef class ControlCurveParameter(BaseControlCurveParameter):
    """ A generic multi-levelled control curve Parameter.

     This parameter can be used to return different values when a `Storage` node's current
      volumes is at different percentage of `max_volume` relative to predefined control curves.
      Control curves must be defined in the range [0, 1] corresponding to 0% and 100% volume.

     By default this parameter returns an integer sequence from zero if the first control curve
      is passed, and incrementing by one for each control curve (or "level") the `Storage` node
      is below.

    Parameters
    ----------
    storage_node : `Storage`
        An optional `Storage` node that can be used to query the current percentage volume.
    control_curves : `float`, `int` or `Parameter` object, or iterable thereof
        The position of the control curves. Internally `float` or `int` types are cast to
        `ConstantParameter`. Multiple values correspond to multiple control curve positions.
        These should be specified in descending order.
    values : array_like or `None`, optional
        The values to return if the `Storage` object is above the correspond control curve.
        I.e. the first value is returned if the current volume is above the first control curve,
        and second value if above the second control curve, and so on. The length of `values`
        must be one more than than the length of `control_curves`.
    parameters : iterable `Parameter` objects or `None`, optional
        If `values` is `None` then `parameters` can specify a `Parameter` object to use at level
        of the control curves. In the same way as `values` the first `Parameter` is used if
        `Storage` is above the first control curve, and second `Parameter` if above the
        second control curve, and so on.
    variable_indices : iterable of ints, optional
        A list of indices that correspond to items in `values` which are to be considered variables
         when `self.is_variable` is True. This mechanism allows a subset of `values` to be variable.
    lower_bounds, upper_bounds : array_like, optional
        Bounds of the variables. The length must correspond to the length of `variable_indices`, i.e.
         there are bounds for each index to be considered as a variable.

    Notes
    -----
    If `values` and `parameters` are both `None`, the default, then `values` defaults to
     a range of integers, starting at zero, one more than length of `control_curves`.

    See also
    --------
    `BaseControlCurveParameter`

    """
    def __init__(self, model, storage_node, control_curves, values=None, parameters=None,
                 variable_indices=None, upper_bounds=None, lower_bounds=None, **kwargs):
        super(ControlCurveParameter, self).__init__(model, storage_node, control_curves, **kwargs)
        # Expected number of values is number of control curves plus one.
        nvalues = len(self.control_curves) + 1
        self.parameters = None
        if values is not None:
            if len(values) != nvalues:
                raise ValueError('Length of values should be one more than the number of '
                                 'control curves ({}).'.format(nvalues))
            self.values = values
        elif parameters is not None:
            if len(parameters) != nvalues:
                raise ValueError('Length of parameters should be one more than the number of '
                                 'control curves ({}).'.format(nvalues))
            self.parameters = list(parameters)
            # Make sure these parameters depend on this parameter to ensure they are evaluated
            # in the correct order.
            for p in self.parameters:
                p.parents.add(self)
        else:
            # No values or parameters given, default to sequence of integers
            self.values = np.arange(nvalues)

        # Default values
        self._upper_bounds = None
        self._lower_bounds = None

        if variable_indices is not None:
            self.variable_indices = variable_indices
            self.double_size = len(variable_indices)
        else:
            self.double_size = 0
        # Bounds for use as a variable (i.e. when self.is_variable = True)
        if upper_bounds is not None:
            if self.values is None or variable_indices is None:
                raise ValueError('Upper bounds can only be specified if `values` and `variable_indices` '
                                 'is not `None`.')
            if len(upper_bounds) != self.double_size:
                raise ValueError('Length of upper_bounds should be equal to the length of `variable_indices` '
                                 '({}).'.format(self.double_size))
            self._upper_bounds = np.array(upper_bounds)

        if lower_bounds is not None:
            if self.values is None or variable_indices is None:
                raise ValueError('Lower bounds can only be specified if `values` and `variable_indices` '
                                 'is not `None`.')
            if len(lower_bounds) != self.double_size:
                raise ValueError('Length of lower_bounds should be equal to the length of `variable_indices` '
                                 '({}).'.format(self.double_size))
            self._lower_bounds = np.array(lower_bounds)

    property values:
        def __get__(self):
            return np.asarray(self._values)
        def __set__(self, values):
            self._values = np.asarray(values, dtype=np.float64)

    property variable_indices:
        def __get__(self):
            return np.array(self._variable_indices)
        def __set__(self, values):
            self._variable_indices = np.array(values, dtype=np.int32)

    @classmethod
    def load(cls, model, data):
        control_curves = super(ControlCurveParameter, cls)._load_control_curves(model, data)
        storage_node = super(ControlCurveParameter, cls)._load_storage_node(model, data)

        parameters = None
        values = None
        if 'values' in data:
            values = load_parameter_values(model, data)
        elif 'parameters' in data:
            # Load parameters
            parameters_data = data['parameters']
            parameters = []
            for pdata in parameters_data:
                parameters.append(load_parameter(model, pdata))

        return cls(model, storage_node, control_curves, values=values, parameters=parameters)

    cpdef double value(self, Timestep ts, ScenarioIndex scenario_index) except? -1:
        cdef int i = scenario_index.global_id
        cdef int j
        cdef AbstractStorage node
        cdef double cc
        cdef Parameter param, cc_param
        node = self.node if self.storage_node is None else self.storage_node

        # Assumes control_curves is sorted highest to lowest
        for j, cc_param in enumerate(self.control_curves):
            cc = cc_param.get_value(scenario_index)
            # If level above control curve then return this level's value
            if node._current_pc[i] >= cc:
                if self.parameters is not None:
                    param = self.parameters[j]
                    return param.get_value(scenario_index)
                else:
                    return self._values[j]

        if self.parameters is not None:
            param = self.parameters[-1]
            return param.get_value(scenario_index)
        else:
            return self._values[-1]

    cpdef set_double_variables(self, double[:] values):
        cdef int i
        cdef double v

        if len(values) != len(self.variable_indices):
            raise ValueError('Number of values must be the same as the number of variable_indices.')

        if self.double_size != 0:
            for i, v in zip(self.variable_indices, values):
                self._values[i] = v

    cpdef double[:] get_double_variables(self):
        cdef int i, j
        cdef double v
        cdef double[:] arry = np.empty((len(self.variable_indices), ))
        for i, j in enumerate(self.variable_indices):
            arry[i] = self._values[j]
        return arry

    cpdef double[:] get_double_lower_bounds(self):
        return self._lower_bounds

    cpdef double[:] get_double_upper_bounds(self):
        return self._upper_bounds

ControlCurveParameter.register()



from pywr.domains.groundwater import KeatingAquifer
from pywr.parameters.groundwater import KeatingStreamFlowParameter
from pywr.core import Model, Input, Output, Link
from pywr.recorders import (NumpyArrayNodeRecorder, NumpyArrayLevelRecorder,
                            NumpyArrayStorageRecorder)
import pandas
import numpy as np
import pytest

num_streams = 1
num_additional_inputs = 1
stream_flow_levels = [[100.0, 125.0]] # m
transmissivity = [1000, 20000] # m2/d
transmissivity = [t * 0.001 for t in transmissivity] # m3 to Ml
coefficient = 1 # no units
storativity = [0.05] # %
levels = [0.0, 1000.0] # m
area = 50000 * 50000 # m2

def test_keating_aquifer():
    model = Model(
        start=pandas.to_datetime('2016-01-01'),
        end=pandas.to_datetime('2016-01-01'),
    )

    aqfer = KeatingAquifer(
        model,
        'keating',
        num_streams,
        num_additional_inputs,
        stream_flow_levels,
        transmissivity,
        coefficient,
        levels,
        area=area,
        storativity=storativity,
    )

    catchment = Input(model, 'catchment', max_flow=0)
    stream = Output(model, 'stream', max_flow=np.inf, cost=0)
    abstraction = Output(model, 'abstraction', max_flow=15, cost=-999)

    catchment.connect(aqfer)
    aqfer.connect(stream, from_slot=0)
    aqfer.connect(abstraction, from_slot=1)

    rec_level = NumpyArrayLevelRecorder(model, aqfer)
    rec_volume = NumpyArrayStorageRecorder(model, aqfer)
    rec_stream = NumpyArrayNodeRecorder(model, stream)
    rec_abstraction = NumpyArrayNodeRecorder(model, abstraction)

    model.check()

    assert(len(aqfer.inputs) == (num_streams + num_additional_inputs))

    for initial_level in (50, 100, 110, 150):
        # set the inital aquifer level and therefor the initial volume
        aqfer.initial_level = initial_level
        initial_volume = aqfer.initial_volume
        assert(initial_volume == (area * storativity[0] * initial_level * 0.001))
        # run the model (for one timestep only)
        model.run()
        # manually calculate keating streamflow and check model flows are OK
        Qp = 2 * transmissivity[0] * max(initial_level - stream_flow_levels[0][0], 0) * coefficient
        Qe = 2 * transmissivity[1] * max(initial_level - stream_flow_levels[0][1], 0) * coefficient
        delta_storage = initial_volume - rec_volume.data[0, 0]
        abs_flow = rec_abstraction.data[0, 0]
        stream_flow = rec_stream.data[0, 0]
        assert(delta_storage == (stream_flow + abs_flow))
        assert(stream_flow == (Qp+Qe))

    A_VERY_LARGE_NUMBER = 9999999999999
    model.timestepper.end = pandas.to_datetime('2016-01-02')

    # fill the aquifer completely
    # there is no spill for the storage so it should find no feasible solution
    with pytest.raises(RuntimeError):
        catchment.max_flow = A_VERY_LARGE_NUMBER
        catchment.min_flow = A_VERY_LARGE_NUMBER
        model.run()

    # drain the aquifer completely
    catchment.min_flow = 0
    catchment.max_flow = 0
    abstraction.max_flow = A_VERY_LARGE_NUMBER
    model.setup()
    model.run()
    assert(rec_volume.data[1, 0] == 0)
    abs_flow = rec_abstraction.data[1, 0]
    stream_flow = rec_stream.data[1, 0]
    assert(stream_flow == 0)
    assert(abs_flow == 0)



# -*- coding: utf-8 -*-

import pywr
from pywr.core import Model, Input, Output, Link, Storage, AggregatedStorage, AggregatedNode

import pandas
import datetime
from helpers import load_model
import pytest


@pytest.fixture()
def model(request):
    model = Model()
    return model


@pytest.fixture()
def simple_linear_model(request):
    """
    Make a simple model with a single Input and Output.

    Input -> Link -> Output

    """
    model = Model()
    inpt = Input(model, name="Input")
    lnk = Link(model, name="Link", cost=1.0)
    inpt.connect(lnk)
    otpt = Output(model, name="Output")
    lnk.connect(otpt)

    return model


@pytest.fixture()
def simple_storage_model(request):
    """
    Make a simple model with a single Input, Storage and Output.

    Input -> Storage -> Output
    """

    model = pywr.core.Model(
        start=pandas.to_datetime('2016-01-01'),
        end=pandas.to_datetime('2016-01-05'),
        timestep=datetime.timedelta(1),
    )

    inpt = Input(model, name="Input", max_flow=5.0, cost=-1)
    res = Storage(model, name="Storage", outputs=1, inputs=1, max_volume=20, initial_volume=10)
    otpt = Output(model, name="Output", max_flow=8, cost=-999)

    inpt.connect(res)
    res.connect(otpt)

    return model


@pytest.fixture()
def three_storage_model(request):
    """
    Make a simple model with three input, storage and output nodes. Also adds
    an `AggregatedStorage` and `AggregatedNode`.

        Input 0 -> Storage 0 -> Output 0
        Input 1 -> Storage 1 -> Output 1
        Input 2 -> Storage 2 -> Output 2


    """
    return load_model("three_storage.json")



# -*- coding: utf-8 -*-
"""
A series of test based on an analytical solution to simple
network problem.


"""
import pywr.core
import datetime
import numpy as np
import pytest
from helpers import assert_model
from fixtures import simple_linear_model

import pywr.parameters


@pytest.mark.parametrize("in_flow, out_flow, benefit",
                         [(10.0, 10.0, 10.0), (10.0, 0.0, 0.0)])
def test_linear_model(simple_linear_model, in_flow, out_flow, benefit):
    """
    Test the simple_linear_model with different basic input and output values
    """

    simple_linear_model.nodes["Input"].max_flow = in_flow
    simple_linear_model.nodes["Output"].min_flow = out_flow
    simple_linear_model.nodes["Output"].cost = -benefit

    expected_sent = in_flow if benefit > 1.0 else out_flow

    expected_node_results = {
        "Input": expected_sent,
        "Link": expected_sent,
        "Output": expected_sent,
    }
    assert_model(simple_linear_model, expected_node_results)


@pytest.fixture(params=[
    (10.0, 5.0, 5.0, 0.0, 0.0, 0.0),
    (10.0, 5.0, 5.0, 0.0, 10.0, 0.0),
    (10.0, 5.0, 5.0, 0.0, 10.0, 2.0),
    (10.0, 5.0, 0.0, 5.0, 10.0, 2.0),
    ])
def linear_model_with_storage(request):
    """
    Make a simple model with a single Input and Output and an offline Storage Node

    Input -> Link -> Output
               |     ^
               v     |
               Storage
    """
    in_flow, out_flow, out_benefit, strg_benefit, current_volume, min_volume = request.param
    max_strg_out = 10.0
    max_volume = 10.0

    model = pywr.core.Model()
    inpt = pywr.core.Input(model, name="Input", min_flow=in_flow, max_flow=in_flow)
    lnk = pywr.core.Link(model, name="Link", cost=0.1)
    inpt.connect(lnk)
    otpt = pywr.core.Output(model, name="Output", min_flow=out_flow, cost=-out_benefit)
    lnk.connect(otpt)

    strg = pywr.core.Storage(model, name="Storage", max_volume=max_volume, min_volume=min_volume,
                             initial_volume=current_volume, cost=-strg_benefit)

    strg.connect(otpt)
    lnk.connect(strg)
    avail_volume = max(current_volume - min_volume, 0.0)
    avail_refill = max_volume - current_volume
    expected_sent = in_flow+min(max_strg_out, avail_volume) if out_benefit > strg_benefit else max(out_flow, in_flow-avail_refill)

    expected_node_results = {
        "Input": in_flow,
        "Link": in_flow,
        "Output": expected_sent,
        "Storage Output": 0.0,
        "Storage Input": min(max_strg_out, avail_volume) if out_benefit > 1.0 else 0.0,
        "Storage": min_volume if out_benefit > strg_benefit else max_volume,
    }
    return model, expected_node_results

def test_linear_model_with_storage(linear_model_with_storage):
    assert_model(*linear_model_with_storage)

@pytest.fixture
def two_domain_linear_model(request):
    """
    Make a simple model with two domains, each with a single Input and Output

    Input -> Link -> Output  : river
                        | across the domain
    Output <- Link <- Input  : grid

    """
    river_flow = 864.0  # Ml/d
    power_plant_cap = 24  # GWh/d
    power_plant_flow_req = 18.0  # Ml/GWh
    power_demand = 12  # GWh/d
    power_benefit = 10.0  # £/GWh

    river_domain = pywr.core.Domain('river')
    grid_domain = pywr.core.Domain('grid')

    model = pywr.core.Model()
    # Create river network
    river_inpt = pywr.core.Input(model, name="Catchment", max_flow=river_flow, domain=river_domain)
    river_lnk = pywr.core.Link(model, name="Reach", domain=river_domain)
    river_inpt.connect(river_lnk)
    river_otpt = pywr.core.Output(model, name="Abstraction", domain=river_domain, cost=0.0)
    river_lnk.connect(river_otpt)
    # Create grid network
    grid_inpt = pywr.core.Input(model, name="Power Plant", max_flow=power_plant_cap, domain=grid_domain,
                                               conversion_factor=1/power_plant_flow_req)
    grid_lnk = pywr.core.Link(model, name="Transmission", cost=1.0, domain=grid_domain)
    grid_inpt.connect(grid_lnk)
    grid_otpt = pywr.core.Output(model, name="Substation", max_flow=power_demand,
                                 cost=-power_benefit, domain=grid_domain)
    grid_lnk.connect(grid_otpt)
    # Connect grid to river
    river_otpt.connect(grid_inpt)

    expected_requested = {'river': 0.0, 'grid': 0.0}
    expected_sent = {'river': power_demand*power_plant_flow_req, 'grid': power_demand}

    expected_node_results = {
        "Catchment": power_demand*power_plant_flow_req,
        "Reach": power_demand*power_plant_flow_req,
        "Abstraction": power_demand*power_plant_flow_req,
        "Power Plant": power_demand,
        "Transmission": power_demand,
        "Substation": power_demand,
    }

    return model, expected_node_results


def test_two_domain_linear_model(two_domain_linear_model):
    assert_model(*two_domain_linear_model)


@pytest.fixture
def two_cross_domain_output_single_input(request):
    """
    Make a simple model with two domains. Thre are two Output nodes
    both connect to an Input node in a different domain.

    In this example the rivers should be able to provide flow to the grid
    with a total flow equal to the sum of their respective parts.

    Input -> Link -> Output  : river
                        | across the domain
                        Input -> Link -> Output : grid
                        | across the domain
    Input -> Link -> Output  : river

    """
    river_flow = 10.0
    expected_node_results = {}

    model = pywr.core.Model()
    # Create grid network
    grid_inpt = pywr.core.Input(model, name="Input", domain='grid',)
    grid_lnk = pywr.core.Link(model, name="Link", cost=1.0, domain='grid')
    grid_inpt.connect(grid_lnk)
    grid_otpt = pywr.core.Output(model, name="Output", max_flow=50.0, cost=-10.0, domain='grid')
    grid_lnk.connect(grid_otpt)
    # Create river network
    for i in range(2):
        river_inpt = pywr.core.Input(model, name="Catchment {}".format(i), max_flow=river_flow, domain='river')
        river_lnk = pywr.core.Link(model, name="Reach {}".format(i), domain='river')
        river_inpt.connect(river_lnk)
        river_otpt = pywr.core.Output(model, name="Abstraction {}".format(i), domain='river', cost=0.0)
        river_lnk.connect(river_otpt)
        # Connect grid to river
        river_otpt.connect(grid_inpt)

        expected_node_results.update({
            "Catchment {}".format(i): river_flow,
            "Reach {}".format(i): river_flow,
            "Abstraction {}".format(i): river_flow
        })

    expected_node_results.update({
        "Input": river_flow*2,
        "Link": river_flow*2,
        "Output": river_flow*2,
    })

    return model, expected_node_results


@pytest.mark.xfail
def test_two_cross_domain_output_single_input(two_cross_domain_output_single_input):
    # TODO This test currently fails because of the simple way in which the cross
    # domain paths work. It can not cope with two Outputs connected to one
    # input.
    assert_model(*two_cross_domain_output_single_input)


@pytest.fixture()
def simple_linear_inline_model(request):
    """
    Make a simple model with a single Input and Output nodes inline of a route.

    Input 0 -> Input 1 -> Link -> Output 0 -> Output 1

    """
    model = pywr.core.Model()
    inpt0 = pywr.core.Input(model, name="Input 0")
    inpt1 = pywr.core.Input(model, name="Input 1")
    inpt0.connect(inpt1)
    lnk = pywr.core.Link(model, name="Link", cost=1.0)
    inpt1.connect(lnk)
    otpt0 = pywr.core.Output(model, name="Output 0")
    lnk.connect(otpt0)
    otpt1 = pywr.core.Output(model, name="Output 1")
    otpt0.connect(otpt1)

    return model


@pytest.mark.skipif(pywr.core.Model().solver.name == "glpk-edge", reason="Not valid for GLPK Edge based solver.")
@pytest.mark.parametrize("in_flow_1, out_flow_0, link_flow",
                         [(10.0, 10.0, 15.0),
                          (0.0, 0.0, 10.0)])
def test_simple_linear_inline_model(simple_linear_inline_model, in_flow_1, out_flow_0, link_flow):
    """
    Test the test_simple_linear_inline_model with different flow constraints
    """
    model = simple_linear_inline_model
    model.nodes["Input 0"].max_flow = 10.0
    model.nodes["Input 1"].max_flow = in_flow_1
    model.nodes["Link"].max_flow = link_flow
    model.nodes["Output 0"].max_flow = out_flow_0
    model.nodes["Input 1"].cost = 1.0
    model.nodes["Output 0"].cost = -10.0
    model.nodes["Output 1"].cost = -5.0

    expected_sent = min(link_flow, 10+in_flow_1)

    expected_node_results = {
        "Input 0": 10.0,
        "Input 1": max(expected_sent-10.0, 0.0),
        "Link": expected_sent,
        "Output 0": min(expected_sent, out_flow_0),
        "Output 1": max(expected_sent - out_flow_0, 0.0),
    }
    assert_model(model, expected_node_results)


@pytest.fixture()
def bidirectional_model(request):
    """
    Make a simple model with a single Input and Output.

    Input 0 -> Link 0 -> Output 0
               |   ^
               v   |
    Input 1 -> Link 1 -> Output 1

    """
    model = pywr.core.Model()
    for i in range(2):
        inpt = pywr.core.Input(model, name="Input {}".format(i))
        lnk = pywr.core.Link(model, name="Link {}".format(i))
        inpt.connect(lnk)
        otpt = pywr.core.Output(model, name="Output {}".format(i))
        lnk.connect(otpt)

    # Create bidirectional link (i.e. a cycle)
    model.nodes['Link 0'].connect(model.nodes['Link 1'])
    model.nodes['Link 1'].connect(model.nodes['Link 0'])

    return model


def test_bidirectional_model(bidirectional_model):
    """
    Test the simple_linear_model with different basic input and output values
    """
    model = bidirectional_model
    model.nodes["Input 0"].max_flow = 10.0
    model.nodes["Input 1"].max_flow = 10.0
    model.nodes["Output 0"].max_flow = 10.0
    model.nodes["Output 1"].max_flow = 15.0
    model.nodes["Output 0"].cost = -5.0
    model.nodes["Output 1"].cost = -10.0
    model.nodes["Link 0"].cost = 1.0
    model.nodes["Link 1"].cost = 1.0

    expected_node_results = {
        "Input 0": 10.0,
        "Input 1": 10.0,
        "Link 0": 10.0,
        "Link 1": 15.0,
        "Output 0": 5.0,
        "Output 1": 15.0,
    }
    assert_model(model, expected_node_results)


def make_simple_model(supply_amplitude, demand, frequency, initial_volume):
    """
    Make a simple model,
        supply -> reservoir -> demand.

    supply is a annual cosine function with amplitude supply_amplitude and
    frequency

    """

    model = pywr.core.Model()

    S = supply_amplitude
    w = frequency

    class SupplyFunc(pywr.parameters.Parameter):
        def value(self, ts, si):
            # Take the mean flow of the day (i.e. offset by half a day)
            t = ts.dayofyear - 0.5
            v = S*np.cos(t*w)+S
            return v

    max_flow = SupplyFunc(model)
    supply = pywr.core.Input(model, name='supply', max_flow=max_flow, min_flow=max_flow)
    demand = pywr.core.Output(model, name='demand', max_flow=demand, cost=-10)
    res = pywr.core.Storage(model, name='reservoir', max_volume=1e6,
                            initial_volume=initial_volume)

    supply_res_link = pywr.core.Link(model, name='link1')
    res_demand_link = pywr.core.Link(model, name='link2')

    supply.connect(supply_res_link)
    supply_res_link.connect(res)
    res.connect(res_demand_link)
    res_demand_link.connect(demand)

    return model


def test_analytical():
    """
    Run the test model though a year with analytical solution values to
    ensure reservoir just contains sufficient volume.
    """

    S = 100.0  # supply amplitude
    D = S  # demand
    w = 2*np.pi/365  # frequency (annual)
    V0 = S/w  # initial reservoir level

    model = make_simple_model(S, D, w, V0)

    T = np.arange(1, 365)
    V_anal = S*(np.sin(w*T)/w+T) - D*T + V0
    V_model = np.empty(T.shape)

    for i, t in enumerate(T):
        model.step()
        V_model[i] = model.nodes['reservoir'].volume[0]

    # Relative error from initial volume
    error = np.abs(V_model - V_anal) / V0
    assert np.all(error < 1e-4)



import pytest
platypus = pytest.importorskip("platypus")
from pywr.optimisation.platypus import PlatypusWrapper, PywrRandomGenerator
from pywr.optimisation import clear_global_model_cache
from platypus import NSGAII, ProcessPoolEvaluator
import os
import numpy as np


TEST_FOLDER = os.path.dirname(__file__)


class TwoReservoirWrapper(PlatypusWrapper):
    def customise_model(self, model):
        self.the_model_has_been_customised = True


@pytest.fixture()
def two_reservoir_problem():
    filename = os.path.join(TEST_FOLDER, 'models', 'two_reservoir.json')
    yield TwoReservoirWrapper(filename)
    # Clean up the
    clear_global_model_cache()
    # We force deallocation the cache here to prevent problems using process pools
    # with pytest.
    import gc
    gc.collect()


@pytest.fixture()
def two_reservoir_constrained_problem():
    filename = os.path.join(TEST_FOLDER, 'models', 'two_reservoir_constrained.json')
    yield TwoReservoirWrapper(filename)
    # Clean up the
    clear_global_model_cache()
    # We force deallocation the cache here to prevent problems using process pools
    # with pytest.
    import gc
    gc.collect()


def test_platypus_init(two_reservoir_problem):
    """ Test the initialisation of the platypus problem. """
    p = two_reservoir_problem

    assert p.problem.nvars == 12
    assert p.problem.nobjs == 2
    assert p.problem.nconstrs == 0
    # Check the `customise_model` method has been called.
    assert p.the_model_has_been_customised


def test_platypus_constrained_init(two_reservoir_constrained_problem):
    """ Test the initialisation of a constrained platypus problem. """
    p = two_reservoir_constrained_problem

    assert p.problem.nvars == 12
    assert p.problem.nobjs == 2
    assert p.problem.nconstrs == 2
    # Check the `customise_model` method has been called.
    assert p.the_model_has_been_customised


def test_platypus_nsgaii_step(two_reservoir_problem):
    """ Undertake a single step of the NSGAII algorithm with a small population. """
    algorithm = NSGAII(two_reservoir_problem.problem, population_size=10)
    algorithm.step()


def test_platypus_nsgaii_step(two_reservoir_constrained_problem):
    """ Undertake a single step of the NSGAII algorithm with a small population. """
    algorithm = NSGAII(two_reservoir_constrained_problem.problem, population_size=10)
    algorithm.step()


def test_platypus_nsgaii_process_pool(two_reservoir_problem):
    """ Undertake a single step of the NSGAII algorithm with a ProcessPool. """
    with ProcessPoolEvaluator(2) as evaluator:
        algorithm = NSGAII(two_reservoir_problem.problem, population_size=50, evaluator=evaluator)
        algorithm.run(10)


class TestPywrRandomGenerator:
    def test_current_model(self, two_reservoir_problem):
        """ Test PywrRandomGenerator inserts the current model configuration in to the population. """
        generator = PywrRandomGenerator(wrapper=two_reservoir_problem)
        algorithm = NSGAII(two_reservoir_problem.problem, population_size=10, generator=generator)
        algorithm.initialize()
        # Ensure the first solution in the population has variable values from the model
        solution = algorithm.population[0]
        np.testing.assert_allclose(solution.variables, np.zeros(12))

    @pytest.mark.parametrize('use_current', [True, False])
    def test_other_solutions(self, two_reservoir_problem, use_current):
        """Test PywrRandomGenerator inserts other solutions into the population. """

        # Create some alternative initial solutions
        solutions = [
            {
                'control_curve': {'doubles': [1] * 12}
            },
            {
                'control_curve': {'doubles': [2] * 12}
            },
        ]

        generator = PywrRandomGenerator(wrapper=two_reservoir_problem, solutions=solutions, use_current=use_current)
        algorithm = NSGAII(two_reservoir_problem.problem, population_size=10, generator=generator)
        algorithm.initialize()
        # Ensure the first solution in the population has variable values from the model
        if use_current:
            np.testing.assert_allclose(algorithm.population[0].variables, np.zeros(12))
            np.testing.assert_allclose(algorithm.population[1].variables, np.ones(12))
            np.testing.assert_allclose(algorithm.population[2].variables, np.ones(12) * 2)
        else:
            np.testing.assert_allclose(algorithm.population[0].variables, np.ones(12))
            np.testing.assert_allclose(algorithm.population[1].variables, np.ones(12) * 2)



from pywr.model import Model
from pywr.nodes import Catchment, Output, DelayNode
from pywr.recorders import NumpyArrayNodeRecorder, AssertionRecorder
from pywr.parameters import ArrayIndexedScenarioParameter, FlowDelayParameter, load_parameter
from pywr.core import Scenario
from numpy.testing import assert_array_almost_equal
import numpy as np
import pytest


@pytest.mark.parametrize("key, delay, initial_flow", [("days", 1, None),
                                                      ("timesteps", 1, None),
                                                      ("days", 1, 5.0),
                                                      ("days", 3, 5.0),
                                                      ("timesteps", 10, 5.0)])
def test_delay_node(key, delay, initial_flow):
    """ Test that the `DelayNode` and the `FlowDelayParameter` internal to it correctly delay node for a range of inputs and
    across scenarios"""
    model = Model()

    model.timestepper.start = "2015/01/01"
    model.timestepper.end = "2015/01/31"

    scen = Scenario(model, name="scenario", size=2)
    flow_vals = np.arange(1, 63).reshape((31, 2), order="F")
    flow = ArrayIndexedScenarioParameter(model, scen, flow_vals)

    catchment = Catchment(model, name="input", flow=flow)
    kwargs = {key: delay}
    if initial_flow:
        kwargs["initial_flow"] = initial_flow
    delaynode = DelayNode(model, name="delaynode", **kwargs)
    output = Output(model, name="output")

    catchment.connect(delaynode)
    delaynode.connect(output)

    rec = NumpyArrayNodeRecorder(model, output)

    model.run()
    if initial_flow:
        expected = np.concatenate([np.full((delay, 2), initial_flow), flow_vals[:-delay, :]])
    else:
        expected = np.concatenate([np.zeros((delay, 2)), flow_vals[:-delay, :]])

    assert_array_almost_equal(rec.data, expected)


def test_delay_param_load():
    """Test that the `.load` method of `FlowDelayParameter` works correctly"""

    model = Model()
    model.timestepper.start = "2015/01/01"
    model.timestepper.end = "2015/01/31"
    catchment = Catchment(model, name="input", flow=1)
    output = Output(model, name="output")
    catchment.connect(output)

    data = {"name": "delay",
            "node": "input",
            "days": 2}

    param = FlowDelayParameter.load(model, data)

    assert param.days == 2

    data = {"name": "delay2",
            "node": "input",
            "timesteps": 2}

    param2 = FlowDelayParameter.load(model, data)

    assert param2.timesteps == 2

    expected = np.concatenate([np.zeros(2), np.ones(29)]).reshape(31, 1)

    AssertionRecorder(model, param, name="rec1", expected_data=expected)
    AssertionRecorder(model, param2, name="rec2", expected_data=expected)

    model.setup()
    model.run()


@pytest.mark.parametrize("key, delay", [("days", 5),
                                        ("timesteps", 0.5)])
def test_delay_failure(key, delay):
    """Test the FlowDelayParameter returns a ValueError when the input value of the `days` attribute is not
    divisible exactly by the model timestep delta and when the `timesteps` attribute is less than 1
    """

    model = Model()
    model.timestepper.start = "2015/01/01"
    model.timestepper.end = "2015/01/31"
    model.timestepper.delta = 3

    catchment = Catchment(model, name="input", flow=1)
    output = Output(model, name="output")
    catchment.connect(output)

    FlowDelayParameter(model, catchment, **{key: delay})

    with pytest.raises(ValueError):
        model.setup()



import logging


logger = logging.getLogger(__name__)


ROOT_NODE = "root"


class GraphInterface:
    def __init__(self, obj):
        self.obj = obj

    @property
    def graph(self):
        return self.obj.model.component_graph

    def add(self, item):
        if not isinstance(item, Component):
            return
        if self is self.obj.children:
            self.graph.add_edge(item, self.obj)
        else:
            self.graph.add_edge(self.obj, item)

    def remove(self, item):
        if not isinstance(item, Component):
            return
        if self is self.obj.children:
            self.graph.remove_edge(item, self.obj)
        else:
            self.graph.remove_edge(self.obj, item)

    def clear(self):
        for n in self._members:
            if self is self.obj.children:
                self.graph.remove_edge(n, self.obj)
            else:
                self.graph.remove_edge(self.obj, n)

    @property
    def _members(self):
        if self is self.obj.children:
            return [n for n in self.graph.predecessors(self.obj) if n != ROOT_NODE]
        else:
            return [x for x in self.graph.successors(self.obj)]

    def __len__(self):
        """Returns the number of nodes in the model"""
        return len(self._members)

    def __iter__(self):
        return iter(self._members)


cdef class Component:
    """ Components of a Model

    This is the base class for all the elements of a `pywr.Model`,
     except the the nodes, that require updates via the `setup`, `reset`,
     `before`, `after` and `finish` methods. This class handles
     registering the instances on the `Model.component_graph` and
     managing the parent/children interface.

    The parent/children interface, through the `pywr.Model.component_graph`
     is used to create a dependency tree such that the methods are
     called in the correct order. E.g. that a `before` method in
     one component that is a parent of another is called first.

    Parameters
    ==========
    name : str or None
        The name of the component.
    comment : str or None
        An optional comment for the component.
    tags : dict (default=None)
        An optional container of key-value pairs that the user can set to help group and identify components.

    See also
    --------
    pywr.Model

    """
    def __init__(self, model, name=None, comment=None, tags=None):
        self.model = model
        self.name = name
        self.comment = comment
        self.tags = tags
        model.component_graph.add_edge(ROOT_NODE, self)
        self.parents = GraphInterface(self)
        self.children = GraphInterface(self)

    property name:
        def __get__(self):
            return self._name

        def __set__(self, name):
            # check for name collision
            if name is not None and name in self.model.components.keys():
                raise ValueError('A component with the name "{}" already exists.'.format(name))
            # apply new name
            self._name = name

    cpdef setup(self):
        logger.debug('Setting up {}: "{}"'.format(self.__class__.__name__, self.name))

    cpdef reset(self):
        logger.debug('Resetting up {}: "{}"'.format(self.__class__.__name__, self.name))

    cpdef before(self):
        pass

    cpdef after(self):
        pass

    cpdef finish(self):
        pass



import os
from pywr.model import Model


def pytest_report_header(config):
    headers = []
    solver_name = Model().solver.name
    headers.append('solver: {}'.format(solver_name))
    return '\n'.join(headers)



from functools import partial
from pywr.recorders import Aggregator
from pywr.recorders._recorders import _agg_func_lookup
from scipy.stats import percentileofscore
import numpy as np
import pytest

npy_funcs = {
    "min": np.min,
    "max": np.max,
    "mean": np.mean,
    "sum": np.sum,
    "median": np.median,
    "product": np.product,
    "all": np.all,
    "any": np.any,
    "count_nonzero": np.count_nonzero
}


def custom_test_func(array, axis=None):
    return np.sum(array**2, axis=axis)


def percentileofscore_with_axis(values, *args, axis=0, **kwargs):
    if values.ndim == 1:
        # For 1D data we just calculate the percentile
        out = percentileofscore(values, *args, **kwargs)
    elif axis == 0:
        # 2D data by axis 0
        out = [percentileofscore(values[:, i], *args, **kwargs) for i in range(values.shape[1])]
    elif axis == 1:
        # 2D data by axis 1
        out = [percentileofscore(values[i, :], *args, **kwargs) for i in range(values.shape[0])]
    else:
        raise ValueError('Axis "{}" not supported'.format(axis))
    return out


@pytest.fixture(params=_agg_func_lookup.keys())
def agg_func(request):
    agg_func_name = request.param

    if agg_func_name == "custom":
        # When using custom you assign the function rather than a string.
        agg_func_name = npy_func = custom_test_func
    elif agg_func_name == "percentile":
        agg_func_name = {
            "func": "percentile",
            "args": [95],
            "kwargs": {}
        }
        npy_func = partial(np.percentile, q=95)
    elif agg_func_name == "percentileofscore":
        agg_func_name = {
            "func": "percentileofscore",
            "kwargs": {
                "score": 0.5,
                "kind": "rank"
            }
        }
        npy_func = partial(percentileofscore_with_axis, score=0.5, kind="rank")
    else:
        npy_func = npy_funcs[agg_func_name]
    return agg_func_name, npy_func


def test_get_set_aggregator(agg_func):
    """Test getter and setter for Aggregator.func"""
    agg_func_name, _ = agg_func
    agg = Aggregator(agg_func_name)
    if isinstance(agg_func_name, dict):
        assert agg.func == agg_func_name['func']
    else:
        assert agg.func == agg_func_name
    agg.func = "sum"
    assert agg.func == "sum"


def test_aggregator_1d(agg_func):
    """ Test Aggregator.aggregate_1d function. """
    agg_func_name, npy_func = agg_func

    agg = Aggregator(agg_func_name)

    data = np.random.rand(10)

    np.testing.assert_allclose(agg.aggregate_1d(data), npy_func(data))


@pytest.mark.parametrize('axis', [0, 1])
def test_aggregator_2d(agg_func, axis):
    """ Test Aggregator.aggregate_2d function. """
    agg_func_name, npy_func = agg_func

    agg = Aggregator(agg_func_name)

    data = np.random.rand(100, 10)

    result = agg.aggregate_2d(data, axis=axis)
    assert len(result) == data.shape[1 - axis]
    np.testing.assert_allclose(result, npy_func(data, axis=axis))



import pytest
pygmo = pytest.importorskip("pygmo")

import pygmo as pg
from pywr.optimisation.pygmo import PygmoWrapper
from pywr.optimisation import clear_global_model_cache
import os


TEST_FOLDER = os.path.dirname(__file__)


@pytest.fixture()
def two_reservoir_wrapper():
    """ Two reservoir test optimisation problem in PygmoWrapper. """
    filename = os.path.join(TEST_FOLDER, 'models', 'two_reservoir.json')
    yield PygmoWrapper(filename)
    # Clean up the
    clear_global_model_cache()


@pytest.fixture()
def two_reservoir_constrained_wrapper():
    """ Two reservoir test optimisation problem in PygmoWrapper. """
    filename = os.path.join(TEST_FOLDER, 'models', 'two_reservoir_constrained.json')
    yield PygmoWrapper(filename)
    # Clean up the
    clear_global_model_cache()


def test_pygmo_single_generation(two_reservoir_wrapper):
    """ Simple pygmo wrapper test. """
    wrapper = two_reservoir_wrapper
    prob = pg.problem(wrapper)
    algo = pg.algorithm(pg.moead(gen=1))

    pg.mp_island.init_pool(2)
    isl = pg.island(algo=algo, prob=prob, size=50, udi=pg.mp_island())
    isl.evolve(1)


def test_pygmo_single_generation_constrained(two_reservoir_constrained_wrapper):
    """ Simple pygmo wrapper test of constrained problem. """
    wrapper = two_reservoir_constrained_wrapper
    prob = pg.problem(wrapper)
    algo = pg.algorithm(pg.moead(gen=1))

    pg.mp_island.init_pool(2)
    isl = pg.island(algo=algo, prob=prob, size=50, udi=pg.mp_island())
    isl.evolve(1)



"""
A collection of tests for pywr.domains.river

Specific additional functionality of the 'special' classes in the river domain
are tested here.
"""
import pywr.core
from pywr.core import Model, Input, Output, Catchment
from pywr.parameters import MonthlyProfileParameter
from pywr.domains import river
from numpy.testing import assert_allclose
import pytest

from helpers import assert_model, load_model

@pytest.fixture(params=[(10.0, 10.0, 10.0), (5.0, 5.0, 1.0)])
def simple_gauge_model(request):
    """
    Make a simple model with a single Input and Output and RiverGauge

    Input -> PiecewiseLink -> Output

    """
    in_flow, out_flow, benefit = request.param
    min_flow_req = 5.0

    model = pywr.core.Model()
    inpt = river.Catchment(model, name="Catchment", flow=in_flow)
    lnk = river.RiverGauge(model, name="Gauge", mrf=min_flow_req, mrf_cost=-1.0)
    inpt.connect(lnk)
    otpt = pywr.core.Output(model, name="Demand", max_flow=out_flow, cost=-benefit)
    lnk.connect(otpt)

    default = inpt.domain

    expected_sent = in_flow if benefit > 1.0 else out_flow

    expected_node_results = {
        "Catchment": expected_sent,
        "Gauge": expected_sent,
        "Gauge Sublink 0": min(min_flow_req, expected_sent),
        "Gauge Sublink 1": expected_sent - min(min_flow_req, expected_sent),
        "Demand": expected_sent,
    }
    return model, expected_node_results


@pytest.fixture
def simple_river_split_gauge_model():
    """
    Make a simple model with a single Input and Output and RiverGauge

    ::

        Input -> RiverSplit -> Output 1
                     \\ --->--> Output 2

    """
    in_flow = 100.0
    min_flow_req = 40.0
    out_flow = 50.0
    model = pywr.core.Model()

    inpt = river.Catchment(model, name="Catchment", flow=in_flow)
    lnk = river.RiverSplitWithGauge(model, name="Gauge", mrf=min_flow_req, mrf_cost=-100,
                                    slot_names=("river", "abstraction"), factors=[3, 1])
    inpt.connect(lnk)
    estuary = pywr.core.Output(model, name="Estuary")
    lnk.connect(estuary, from_slot="river")
    otpt = pywr.core.Output(model, name="Demand", max_flow=out_flow, cost=-10)
    lnk.connect(otpt, from_slot="abstraction")

    net_flow_after_mrf = in_flow - min_flow_req
    expected_node_results = {
        "Catchment": in_flow,
        "Gauge": in_flow,
        "Gauge Sublink 0": min_flow_req,
        "Gauge Sublink 1": net_flow_after_mrf * 0.75,
        "Gauge Sublink 2": net_flow_after_mrf * 0.25,
        "Demand": min(out_flow, net_flow_after_mrf * 0.25),
    }
    return model, expected_node_results


def test_river_gauge():
    """
    Test loading a model with a RiverGauge from JSON, modifying it, then running it
    """
    model = load_model("river_mrf1.json")

    node = model.nodes["mrf"]
    demand = model.nodes["demand"]

    # test getting properties
    assert(isinstance(node.mrf, MonthlyProfileParameter))
    assert_allclose(node.mrf_cost, -1000)
    assert_allclose(node.cost, 0.0)

    # test setting properties
    node.mrf = 40
    node.mrf_cost = -999
    assert_allclose(node.mrf, 40)
    assert_allclose(node.mrf_cost, -999)

    # run the model and see if it works
    model.run()
    assert_allclose(node.flow, 40)
    assert_allclose(demand.flow, 60)


def test_piecewise_model(simple_gauge_model):
    assert_model(*simple_gauge_model)


def test_river_split_gauge(simple_river_split_gauge_model):
    assert_model(*simple_river_split_gauge_model)

def test_river_split_gauge_json():
    """As test_river_split_gauge, but model is defined in JSON"""

    model = load_model("river_split_with_gauge1.json")
    model.check()
    model.run()

    catchment_flow = 100.0
    mrf = 40.0
    expected_demand_flow = (catchment_flow - mrf) * 0.25

    catchment_node = model.nodes["Catchment"]
    demand_node = model.nodes["Demand"]
    assert_allclose(catchment_node.flow, catchment_flow)
    assert_allclose(demand_node.flow, expected_demand_flow)


def test_control_curve():
    """
    Use a simple model of a Reservoir to test that a control curve
    behaves as expected.

    The control curve should alter the cost of the Reservoir when it
    is above or below a particular threshold.

    (flow = 8.0)          (max_flow = 10.0)
    Catchment -> River -> DemandCentre
                     |        ^
    (max_flow = 2.0) v        | (max_flow = 2.0)
                    Reservoir

    """
    in_flow = 8

    model = pywr.core.Model()
    catchment = river.Catchment(model, name="Catchment", flow=in_flow)
    lnk = river.River(model, name="River")
    catchment.connect(lnk)
    demand = pywr.core.Output(model, name="Demand", cost=-10.0, max_flow=10)
    lnk.connect(demand)
    from pywr.parameters import ConstantParameter
    control_curve = ConstantParameter(model, 0.8)
    reservoir = river.Reservoir(model, name="Reservoir", max_volume=10, cost=-20, above_curve_cost=0.0,
                                control_curve=control_curve, initial_volume=10)
    reservoir.inputs[0].max_flow = 2.0
    reservoir.outputs[0].max_flow = 2.0
    lnk.connect(reservoir)
    reservoir.connect(demand)

    model.setup()

    model.step()
    # Reservoir is currently above control curve. 2 should be taken from the
    # reservoir
    assert(reservoir.volume == 8)
    assert(demand.flow == 10)
    # Reservoir is still at (therefore above) control curve. So 2 is still taken
    model.step()
    assert(reservoir.volume == 6)
    assert(demand.flow == 10)
    # Reservoir now below curve. Better to retain volume and divert some of the
    # inflow
    model.step()
    assert(reservoir.volume == 8)
    assert(demand.flow == 6)
    # Set the above_curve_cost function to keep filling
    from pywr.parameters.control_curves import ControlCurveParameter
    # We know what we're doing with the control_curve Parameter so unset its parent before overriding
    # the cost parameter.
    # We need to call setup() again because we're adding a parameter.
    reservoir.cost = ControlCurveParameter(model, reservoir, control_curve, [-20.0, -20.0])
    reservoir.initial_volume = 8
    model.setup()
    model.step()
    assert(reservoir.volume == 10)
    assert(demand.flow == 6)

def test_catchment_many_successors():
    """Test if node with fixed flow can have multiple successors. See #225"""
    model = Model()
    catchment = Catchment(model, "catchment", flow=100)
    out1 = Output(model, "out1", max_flow=10, cost=-100)
    out2 = Output(model, "out2", max_flow=15, cost=-50)
    out3 = Output(model, "out3")
    catchment.connect(out1)
    catchment.connect(out2)
    catchment.connect(out3)
    model.check()
    model.run()
    assert_allclose(out1.flow, 10)
    assert_allclose(out2.flow, 15)
    assert_allclose(out3.flow, 75)



import numpy as np
from pywr import _core
from pywr._core import Node as BaseNode
from pywr._core import BaseInput, BaseLink, BaseOutput, StorageInput, StorageOutput
from pywr.parameters import pop_kwarg_parameter, load_parameter, load_parameter_values, FlowDelayParameter
from pywr.domains import Domain
import networkx as nx


class Drawable(object):
    """Mixin class for objects that are drawable on a diagram of the network.
    """
    def __init__(self, *args, **kwargs):
        self.position = kwargs.pop('position', {})
        self.color = kwargs.pop('color', 'black')
        self.visible = kwargs.pop('visible', True)
        super(Drawable, self).__init__(*args, **kwargs)


class Loadable:
    """Mixin class that laods nodes from JSON data.

    Loading is performed in two stages. First a `pre_load` classmethod creates a node instance passing
    any non-parameter arguments via keyword to the respective Node's init method. After parameters are
    loaded (see `Model.load`) the `finalise_load` method is called on all node instances to assign
    concrete parameter instances where needed.
    """
    __parameter_attributes__ = ()
    __node_attributes__ = ()
    __parameter_value_attributes__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__parameters_to_load = None

    @classmethod
    def _pre_load_parameter(cls, key, value, param_data, non_param_data):
        # By default parameter references are loaded later, and no keyword is given to the `__init__`
        param_data[key] = value

    @classmethod
    def pre_load(cls, model, data):
        """Create a node instance from data.

        Parameter data and references are stored until later, and consumed by a call to `finalise_load`.
        """
        # Filter non-parameter data and initialise with it only
        param_data = {}
        non_param_data = {}
        for key, value in data.items():
            if key in cls.__parameter_attributes__:
                cls._pre_load_parameter(key, value, param_data, non_param_data)
            elif key in cls.__node_attributes__:
                # Node references are converted to nodes immediately.
                if isinstance(value, list):
                    node = [model.pre_load_node(n) for n in value]
                else:
                    node = model.pre_load_node(value)
                non_param_data[key] = node
            elif key in cls.__parameter_value_attributes__:
                if isinstance(value, (float, int)):
                    non_param_data[key] = value
                else:
                    non_param_data[key] = load_parameter_values(model, value)
            else:
                non_param_data[key] = value

        obj = cls(model=model, **non_param_data)
        obj.__parameters_to_load = param_data
        return obj

    def finalise_load(self):
        """Finish loading a node by converting parameter name references to instance references."""
        for key, parameter_data in self.__parameters_to_load.items():
            if isinstance(parameter_data, list):
                # List of parameter references
                parameter = [load_parameter(self.model, d) for d in parameter_data]
            else:
                parameter = load_parameter(self.model, parameter_data)

            setattr(self, key, parameter)
        del self.__parameters_to_load


class Connectable(object):
    """A mixin class providing methods for connecting nodes in the model graph"""
    def iter_slots(self, slot_name=None, is_connector=True):
        """ Returns the object(s) wich should be connected to given slot_name

        Overload this method when implementing compound nodes which have
        multiple slots and may return something other than self.

        is_connector is True when self's connect method has been used. I.e. self
        is connecting to another object. This is useful for providing an
        appropriate response object in circumstances where a subnode should make
        the actual connection rather than self.
        """
        if slot_name is not None:
            raise ValueError('{} does not have slot: {}'.format(self, slot_name))
        yield self

    def connect(self, node, from_slot=None, to_slot=None):
        """Create an edge from this Node to another Node

        Parameters
        ----------
        node : Node
            The node to connect to
        from_slot : object (optional)
            The outgoing slot on this node to connect to
        to_slot : object (optional)
            The incoming slot on the target node to connect to
        """
        if self.model is not node.model:
            raise RuntimeError("Can't connect Nodes in different Models")
        if not isinstance(node, Connectable):
            raise TypeError("Other node ({}) is not connectable.".format(node))

        # Get slot from this node
        for node1 in self.iter_slots(slot_name=from_slot, is_connector=True):
            # And slot to connect from other node
            for node2 in node.iter_slots(slot_name=to_slot, is_connector=False):
                self.model.graph.add_edge(node1, node2)
        self.model.dirty = True

    def disconnect(self, node=None, slot_name=None, all_slots=True):
        """Remove a connection from this Node to another Node

        Parameters
        ----------
        node : Node (optional)
            The node to remove the connection to. If another node is not
            specified, all connections from this node will be removed.
        slot_name : integer (optional)
            If specified, only remove the connection to a specific slot name.
            Otherwise connections from all slots are removed.
        """
        if node is not None:
            self._disconnect(node, slot_name=slot_name, all_slots=all_slots)
        else:
            neighbors = self.model.graph.neighbors(self)
            for neighbor in [neighbor for neighbor in neighbors]:
                self._disconnect(neighbor, slot_name=slot_name, all_slots=all_slots)

    def _disconnect(self, node, slot_name=None, all_slots=True):
        """As disconnect, except node argument is required"""
        disconnected = False
        try:
            self.model.graph.remove_edge(self, node)
        except nx.exception.NetworkXError:
            for node_slot in node.iter_slots(slot_name=slot_name, is_connector=False, all_slots=all_slots):
                try:
                    self.model.graph.remove_edge(self, node_slot)
                except nx.exception.NetworkXError:
                    pass
                else:
                    disconnected = True
        else:
            disconnected = True
        if not disconnected:
            raise nx.exception.NetworkXError('{} is not connected to {}'.format(self, node))
        self.model.dirty = True


class NodeMeta(type):
    """Node metaclass used to keep a registry of Node classes"""
    # node subclasses are stored in a dict for convenience
    node_registry = {}

    def __new__(meta, name, bases, dct):
        return super(NodeMeta, meta).__new__(meta, name, bases, dct)

    def __init__(cls, name, bases, dct):
        super(NodeMeta, cls).__init__(name, bases, dct)
        cls.node_registry[name.lower()] = cls

    def __call__(cls, *args, **kwargs):
        # Create new instance of Node (or subclass thereof)
        node = type.__call__(cls, *args, **kwargs)
        # Add node to Model graph. This needs to be done here, so that if the
        # __init__ method of Node raises an exception it is not added.
        node.model.graph.add_node(node)
        node.model.dirty = True
        return node


class Node(Loadable, Drawable, Connectable, BaseNode, metaclass=NodeMeta):
    """Base object from which all other nodes inherit

    This BaseNode is not connectable by default, and the Node class should
    be used for actual Nodes in the model. The BaseNode provides an abstract
    class for other Node types (e.g. StorageInput) that are not directly
    Connectable.
    """
    __parameter_attributes__ = ('cost', 'min_flow', 'max_flow')

    def __init__(self, model, name, **kwargs):
        """Initialise a new Node object

        Parameters
        ----------
        model : Model
            The model the node belongs to
        name : string
            A unique name for the node
        """
        color = kwargs.pop('color', 'black')
        min_flow = pop_kwarg_parameter(kwargs, 'min_flow', 0.0)
        if min_flow is None:
            min_flow = 0.0
        max_flow = pop_kwarg_parameter(kwargs, 'max_flow', float('inf'))
        cost = pop_kwarg_parameter(kwargs, 'cost', 0.0)
        conversion_factor = pop_kwarg_parameter(kwargs, 'conversion_factor', 1.0)

        super(Node, self).__init__(model, name, **kwargs)

        self.slots = {}
        self.color = color
        self.min_flow = min_flow
        self.max_flow = max_flow
        self.cost = cost
        self.conversion_factor = conversion_factor

    def check(self):
        """Check the node is valid

        Raises an exception if the node is invalid
        """
        pass


class Input(Node, BaseInput):
    """A general input at any point in the network

    """
    def __init__(self, *args, **kwargs):
        """Initialise a new Input node

        Parameters
        ----------
        min_flow : float (optional)
            A simple minimum flow constraint for the input. Defaults to None
        max_flow : float (optional)
            A simple maximum flow constraint for the input. Defaults to 0.0
        """
        super(Input, self).__init__(*args, **kwargs)
        self.color = '#F26C4F'  # light red


class Output(Node, BaseOutput):
    """A general output at any point from the network

    """
    def __init__(self, *args, **kwargs):
        """Initialise a new Output node

        Parameters
        ----------
        min_flow : float (optional)
            A simple minimum flow constraint for the output. Defaults to 0.0
        max_flow : float (optional)
            A simple maximum flow constraint for the output. Defaults to None
        """
        kwargs['color'] = kwargs.pop('color', '#FFF467')  # light yellow
        super(Output, self).__init__(*args, **kwargs)


class Link(Node, BaseLink):
    """A link in the supply network, such as a pipe

    Connections between Nodes in the network are created using edges (see the
    Node.connect and Node.disconnect methods). However, these edges cannot
    hold constraints (e.g. a maximum flow constraint). In this instance a Link
    node should be used.
    """
    def __init__(self, *args, **kwargs):
        """Initialise a new Link node

        Parameters
        ----------
        max_flow : float or function (optional)
            A maximum flow constraint on the link, e.g. 5.0
        """
        kwargs['color'] = kwargs.pop('color', '#A0A0A0')  # 45% grey
        super(Link, self).__init__(*args, **kwargs)


class Storage(Loadable, Drawable, Connectable, _core.Storage, metaclass=NodeMeta):
    """A generic storage Node

    In terms of connections in the network the Storage node behaves like any
    other node, provided there is only 1 input and 1 output. If there are
    multiple sub-nodes the connections need to be explicit about which they
    are connecting to. For example:

    >>> storage(model, 'reservoir', num_outputs=1, num_inputs=2)
    >>> supply.connect(storage)
    >>> storage.connect(demand1, from_slot=0)
    >>> storage.connect(demand2, from_slot=1)

    The attribtues of the sub-nodes can be modified directly (and
    independently). For example:

    >>> storage.outputs[0].max_flow = 15.0

    If a recorder is set on the storage node, instead of recording flow it
    records changes in storage. Any recorders set on the output or input
    sub-nodes record flow as normal.

    Parameters
    ----------
    model : Model
        Model instance to which this storage node is attached.
    name : str
        The name of the storage node.
    num_inputs, num_outputs : integer (optional)
        The number of input and output nodes to create internally. Defaults to 1.
    min_volume : float (optional)
        The minimum volume of the storage. Defaults to 0.0.
    max_volume : float, Parameter (optional)
        The maximum volume of the storage. Defaults to 0.0.
    initial_volume, initial_volume_pc : float (optional)
        Specify initial volume in either absolute or proportional terms. Both are required if `max_volume`
        is a parameter because the parameter will not be evaluated at the first time-step. If both are given
        and `max_volume` is not a Parameter, then the absolute value is ignored.
    cost : float, Parameter (optional)
        The cost of net flow in to the storage node. I.e. a positive cost penalises increasing volume by
        giving a benefit to negative net flow (release), and a negative cost penalises decreasing volume
        by giving a benefit to positive net flow (inflow).
    area, level : float, Parameter (optional)
        Optional float or Parameter defining the area and level of the storage node. These values are
        accessible through the `get_area` and `get_level` methods respectively.
    """
    __parameter_attributes__ = ('cost', 'min_volume', 'max_volume', 'level', 'area')
    __parameter_value_attributes__ = ('initial_volume', )

    def __init__(self, model, name, outputs=1, inputs=1, *args, **kwargs):

        min_volume = pop_kwarg_parameter(kwargs, 'min_volume', 0.0)
        if min_volume is None:
            min_volume = 0.0
        max_volume = pop_kwarg_parameter(kwargs, 'max_volume', 0.0)
        initial_volume = kwargs.pop('initial_volume', None)
        initial_volume_pc = kwargs.pop('initial_volume_pc', None)
        cost = pop_kwarg_parameter(kwargs, 'cost', 0.0)
        level = pop_kwarg_parameter(kwargs, 'level', None)
        area = pop_kwarg_parameter(kwargs, 'area', None)

        super(Storage, self).__init__(model, name, **kwargs)

        self.outputs = []
        for n in range(0, outputs):
            self.outputs.append(StorageOutput(model, name="[output{}]".format(n), parent=self))

        self.inputs = []
        for n in range(0, inputs):
            self.inputs.append(StorageInput(model, name="[input{}]".format(n), parent=self))

        self.min_volume = min_volume
        self.max_volume = max_volume
        self.initial_volume = initial_volume
        self.initial_volume_pc = initial_volume_pc
        self.cost = cost
        self.level = level
        self.area = area

        # StorageOutput and StorageInput are Cython classes, which do not have
        # NodeMeta as their metaclass, therefore they don't get added to the
        # model graph automatically.
        for node in self.outputs:
            self.model.graph.add_node(node)
        for node in self.inputs:
            self.model.graph.add_node(node)

    def iter_slots(self, slot_name=None, is_connector=True, all_slots=False):
        if is_connector:
            if not self.inputs:
                raise StopIteration
            if slot_name is None:
                if all_slots or len(self.inputs) == 1:
                    for node in self.inputs:
                        yield node
                else:
                    raise ValueError("Must specify slot identifier.")
            else:
                try:
                    yield self.inputs[slot_name]
                except IndexError:
                    raise IndexError('{} does not have slot: {}'.format(self, slot_name))
        else:
            if not self.outputs:
                raise StopIteration
            if slot_name is None:
                if all_slots or len(self.outputs) == 1:
                    for node in self.outputs:
                        yield node
                else:
                    raise ValueError("Must specify slot identifier.")
            else:
                yield self.outputs[slot_name]

    def check(self):
        pass  # TODO

    def __repr__(self):
        return '<{} "{}">'.format(self.__class__.__name__, self.name)


class VirtualStorage(Loadable, Drawable, _core.VirtualStorage, metaclass=NodeMeta):
    """A virtual storage unit

    Parameters
    ----------
    model: pywr.core.Model
    name: str
        The name of the virtual node
    nodes: list of nodes
        List of inflow/outflow nodes that affect the storage volume
    factors: list of floats
        List of factors to multiply node flow by. Positive factors remove
        water from the storage, negative factors remove it.
    min_volume: float or parameter
        The minimum volume the storage is allowed to reach.
    max_volume: float or parameter
        The maximum volume of the storage.
    initial_volume: float
        The initial storage volume.
    cost: float or parameter
        The cost of flow into/outfrom the storage.

    Notes
    -----
    TODO: The cost property is not currently respected. See issue #242.
    """
    __parameter_attributes__ = ('min_volume', 'max_volume')
    __node_attributes__ = ('nodes', )

    def __init__(self, model, name, nodes, **kwargs):
        min_volume = pop_kwarg_parameter(kwargs, 'min_volume', 0.0)
        if min_volume is None:
            min_volume = 0.0
        max_volume = pop_kwarg_parameter(kwargs, 'max_volume', 0.0)
        if 'volume' in kwargs:
            # support older API where volume kwarg was the initial volume
            initial_volume = kwargs.pop('volume')
        else:
            initial_volume = kwargs.pop('initial_volume', 0.0)
        cost = pop_kwarg_parameter(kwargs, 'cost', 0.0)

        factors = kwargs.pop('factors', None)

        super(VirtualStorage, self).__init__(model, name, **kwargs)

        self.min_volume = min_volume
        self.max_volume = max_volume
        self.initial_volume = initial_volume
        self.cost = cost
        self.nodes = nodes

        if factors is None:
            self.factors = [1.0 for i in range(len(nodes))]
        else:
            self.factors = factors

    def check(self):
        super(VirtualStorage, self).check()
        if self.cost not in (0.0, None):
            raise NotImplementedError("VirtualStorage does not currently support a non-zero cost.")


class RollingVirtualStorage(Loadable, Drawable, _core.RollingVirtualStorage, metaclass=NodeMeta):
    """A rolling virtual storage node useful for implementing rolling licences.

    Parameters
    ----------
    model: pywr.core.Model
    name: str
        The name of the virtual node
    nodes: list of nodes
        List of inflow/out flow nodes that affect the storage volume
    factors: list of floats
        List of factors to multiply node flow by. Positive factors remove
        water from the storage, negative factors remove it.
    min_volume: float or parameter
        The minimum volume the storage is allowed to reach.
    max_volume: float or parameter
        The maximum volume of the storage.
    initial_volume: float
        The initial storage volume.
    timesteps : int
        The number of timesteps to apply to the rolling storage over.
    days : int
        The number of days to apply the rolling storage over. Specifying a number of days (instead of a number
        of timesteps) is only valid with models running a timestep of daily frequency.
    cost: float or parameter
        The cost of flow into/outfrom the storage.

    Notes
    -----
    TODO: The cost property is not currently respected. See issue #242.
    """
    __parameter_attributes__ = ('min_volume', 'max_volume')
    __node_attributes__ = ('nodes', )

    def __init__(self, model, name, nodes, **kwargs):
        min_volume = pop_kwarg_parameter(kwargs, 'min_volume', 0.0)
        if min_volume is None:
            min_volume = 0.0
        max_volume = pop_kwarg_parameter(kwargs, 'max_volume', 0.0)
        initial_volume = kwargs.pop('initial_volume', 0.0)
        cost = pop_kwarg_parameter(kwargs, 'cost', 0.0)
        factors = kwargs.pop('factors', None)
        days = kwargs.pop('days', None)
        timesteps = kwargs.pop('timesteps', 0)

        if not timesteps and not days:
            raise ValueError("Either `timesteps` or `days` must be specified.")

        super().__init__(model, name, **kwargs)

        self.min_volume = min_volume
        self.max_volume = max_volume
        self.initial_volume = initial_volume
        self.cost = cost
        self.nodes = nodes
        self.days = days
        self.timesteps = timesteps

        if factors is None:
            self.factors = [1.0 for i in range(len(nodes))]
        else:
            self.factors = factors

    def check(self):
        super().check()
        if self.cost not in (0.0, None):
            raise NotImplementedError("RollingVirtualStorage does not currently support a non-zero cost.")

    def setup(self, model):
        if self.days is not None and self.days > 0:
            try:
                self.timesteps = self.days // self.model.timestepper.delta
            except TypeError:
                raise TypeError('A rolling period defined as a number of days is only valid with daily time-steps.')
        if self.timesteps < 1:
            raise ValueError('The number of time-steps for a RollingVirtualStorage node must be greater than one.')
        super().setup(model)


class AnnualVirtualStorage(VirtualStorage):
    """A virtual storage which resets annually, useful for licences

    See documentation for `pywr.core.VirtualStorage`.

    Parameters
    ----------
    reset_day: int
        The day of the month (0-31) to reset the volume to the initial value.
    reset_month: int
        The month of the year (0-12) to reset the volume to the initial value.
    reset_to_initial_volume: bool
        Reset the volume to the initial volume instead of maximum volume each year (default is False).

    """
    def __init__(self, *args, **kwargs):
        self.reset_day = kwargs.pop('reset_day', 1)
        self.reset_month = kwargs.pop('reset_month', 1)
        self.reset_to_initial_volume = kwargs.pop('reset_to_initial_volume', False)
        self._last_reset_year = None

        super(AnnualVirtualStorage, self).__init__(*args, **kwargs)

    def reset(self):
        super(AnnualVirtualStorage, self).reset()
        self._last_reset_year = None

    def before(self, ts):
        super(AnnualVirtualStorage, self).before(ts)

        # Reset the storage volume if necessary
        if ts.year != self._last_reset_year:
            # I.e. we're in a new year and ...
            # ... we're at or past the reset month/day
            if ts.month > self.reset_month or \
                    (ts.month == self.reset_month and ts.day >= self.reset_day):
                # Reset to maximum volume (i.e. full capacity. )
                self._reset_storage_only(use_initial_volume=self.reset_to_initial_volume)
                self._last_reset_year = ts.year
                self.active = True


class SeasonalVirtualStorage(AnnualVirtualStorage):
    """A virtual storage node that operates only for a specified period within a year.

    This node is most useful for representing licences that are only enforced during specified periods. The
    `reset_day` and `reset_month` parameters indicate when the node starts operating and the `end_day` and `end_month`
    when it stops operating. For the period when the node is not operating, the volume of the node remains unchanged
    and the node does not apply any constraints to the model.

    The end_day and end_month can represent a date earlier in the year that the reset_day and and reset_month. This
    situation represents a licence that operates across a year boundary. For example, one that is active between
    October and March and not active between April and September.

    Parameters
    ----------
    reset_day : int
        The day of the month (0-31) when the node starts operating and its volume is reset to the initial value or
        maximum volume.
    reset_month : int
        The month of the year (0-12) when the node starts operating and its volume is reset to the initial value or
        maximum volume.
    reset_to_initial_volume : bool
        Reset the volume to the initial volume instead of maximum volume each year (default is False).
    end_day : int
        The day of the month (0-31) when the node stops operating.
    end_month : int
        The month of the year (0-12) when the node stops operating.
    """

    def __init__(self, *args, **kwargs):
        self.end_day = kwargs.pop('end_day', 31)
        self.end_month = kwargs.pop('end_month', 12)
        self._last_active_year = None

        super().__init__(*args, **kwargs)

    def before(self, ts):
        super().before(ts)

        if ts.year != self._last_active_year:
            if ts.index == 0:
                if self._last_reset_year == ts.year:
                    # First timestep is later in year than reset date
                    if self.end_month < self.reset_month or \
                            (self.end_month == self.reset_month and self.end_day <= self.reset_day):
                        # end date is earlier in year than reset date so do not deactivate node in first year
                        self._last_active_year = ts.year
                else:
                    # First timestep is earlier in year than reset date
                    if self.end_month > self.reset_month or \
                            (self.end_month == self.reset_month and self.end_day >= self.reset_day):
                        # end date is later in year than reset date so node needs to be deactivated
                        self.active = False
            elif ts.month > self.end_month or \
                    (ts.month == self.end_month and ts.day >= self.end_day):
                self._last_active_year = ts.year
                self.active = False


class PiecewiseLink(Node):
    """ An extension of Node that represents a non-linear Link with a piece wise cost function.

    This object is intended to model situations where there is a benefit of supplying certain flow rates
    but beyond a fixed limit there is a change in (or zero) cost.

    Parameters
    ----------
    max_flows : iterable
        A monotonic increasing list of maximum flows for the piece wise function
    costs : iterable
        A list of costs corresponding to the max_flow steps

    Notes
    -----

    This Node is implemented using a compound node structure like so:

    ::

                | Separate Domain         |
        Output -> Sublink 0 -> Sub Output -> Input
               -> Sublink 1 ---^
               ...             |
               -> Sublink n ---|

    This means routes do not directly traverse this node due to the separate
    domain in the middle. Instead several new routes are made for each of
    the sublinks and connections to the Output/Input node. The reason for this
    breaking of the route is to avoid an geometric increase in the number
    of routes when multiple PiecewiseLinks are present in the same route.
    """
    __parameter_attributes__ = ('costs', 'max_flows')

    def __init__(self, model, nsteps, *args, **kwargs):
        self.allow_isolated = True
        costs = kwargs.pop('costs', None)
        max_flows = kwargs.pop('max_flows', None)


        # TODO look at the application of Domains here. Having to use
        # Input/Output instead of BaseInput/BaseOutput because of a different
        # domain is required on the sub-nodes and they need to be connected
        self.sub_domain = Domain()
        self.input = Input(model, name='{} Input'.format(self.name), parent=self)
        self.output = Output(model, name='{} Output'.format(self.name), parent=self)

        self.sub_output = Output(model, name='{} Sub Output'.format(self.name), parent=self,
                                 domain=self.sub_domain)
        self.sub_output.connect(self.input)
        self.sublinks = []
        for i in range(nsteps):
            sublink = Input(model, name='{} Sublink {}'.format(self.name, i), parent=self, domain=self.sub_domain)
            self.sublinks.append(sublink)
            sublink.connect(self.sub_output)
            self.output.connect(self.sublinks[-1])

        super().__init__(model, *args, **kwargs)

        if costs is not None:
            self.costs = costs
        if max_flows is not None:
            self.max_flows = max_flows

    def costs():
        def fget(self):
            return [sl.cost for sl in self.sublinks]

        def fset(self, values):
            if len(self.sublinks) != len(values):
                raise ValueError(f"Piecewise costs must be the same length as the number of "
                                 f"sub-links ({len(self.sublinks)}).")
            for i, sl in enumerate(self.sublinks):
                sl.cost = values[i]
        return locals()
    costs = property(**costs())

    def max_flows():
        def fget(self):
            return [sl.max_flow for sl in self.sublinks]

        def fset(self, values):
            if len(self.sublinks) != len(values):
                raise ValueError(f"Piecewise max_flows must be the same length as the number of "
                                 f"sub-links ({len(self.sublinks)}).")
            for i, sl in enumerate(self.sublinks):
                sl.max_flow = values[i]
        return locals()
    max_flows = property(**max_flows())

    def iter_slots(self, slot_name=None, is_connector=True):
        if is_connector:
            yield self.input
        else:
            yield self.output

    def after(self, timestep):
        """
        Set total flow on this link as sum of sublinks
        """
        for lnk in self.sublinks:
            self.commit_all(lnk.flow)
        # Make sure save is done after setting aggregated flow
        super(PiecewiseLink, self).after(timestep)


class MultiSplitLink(PiecewiseLink):
    """ An extension of PiecewiseLink that includes additional slots to connect from.

    Conceptually this node looks like the following internally,

    ::

                 / -->-- X0 -->-- \\
        A -->-- Xo -->-- X1 -->-- Xi -->-- C
                 \\ -->-- X2 -->-- /
                         |
                         Bo -->-- Bi --> D

    An additional sublink in the PiecewiseLink (i.e. X2 above) and nodes
    (i.e. Bo and Bi) in this class are added for each extra slot.

    Finally a mechanism is provided to (optionally) fix the ratio between the
    last non-split sublink (i.e. X1) and each of the extra sublinks (i.e. X2).
    This mechanism uses `AggregatedNode` internally.

    Parameters
    ----------
    max_flows : iterable
        A monotonic increasing list of maximum flows for the piece wise function
    costs : iterable
        A list of costs corresponding to the max_flow steps
    extra_slots : int, optional (default 1)
        Number of additional slots (and sublinks) to provide. Must be greater
        than zero.
    slot_names : iterable, optional (default range of ints)
        The names by which to refer to the slots during connection to other
        nodes. Length must be one more than the number of extra_slots. The first
        item refers to the PiecewiseLink connection with the following items for
        each extra slot.
    factors : iterable, optional (default None)
        If given, the length must be equal to one more than the number of
        extra_slots. Each item is the proportion of total flow to pass through
        the additional sublinks. If no factor is required for a particular
        sublink then use `None` for its items. Factors are normalised prior to
        use in the solver.

    Notes
    -----
    Users must be careful when using the factor mechanism. Factors use the last
    non-split sublink (i.e. X1 but not X0). If this link is constrained with a
    maximum or minimum flow, or if it there is another unconstrained link
    (i.e. if X0 is unconstrained) then ratios across this whole node may not be
    enforced as expected.

    """
    def __init__(self, model, nsteps, *args, **kwargs):
        self.allow_isolated = True
        costs = kwargs.pop('costs', None)
        max_flows = kwargs.pop('max_flows', None)

        extra_slots = kwargs.pop('extra_slots', 1)
        if extra_slots < 1:
            raise ValueError("extra_slots must be at least 1.")

        # Default to integer names
        self.slot_names = list(kwargs.pop('slot_names', range(extra_slots+1)))
        if extra_slots+1 != len(self.slot_names):
            raise ValueError("slot_names must be one more than the number of extra_slots.")

        factors = kwargs.pop('factors', None)
        # Finally initialise the parent.
        super(MultiSplitLink, self).__init__(model, nsteps + extra_slots, *args, **kwargs)

        self._extra_inputs = []
        self._extra_outputs = []
        n = len(self.sublinks) - extra_slots
        for i in range(extra_slots):
            # create a new input inside the piecewise link which only has access
            # to flow travelling via the last sublink (X2)
            otpt = Output(self.model, '{} Extra Output {}'.format(self.name, i),
                          domain=self.sub_domain, parent=self)
            inpt = Input(self.model, '{} Extra Input {}'.format(self.name, i), parent=self)

            otpt.connect(inpt)
            self.sublinks[n+i].connect(otpt)

            self._extra_inputs.append(inpt)
            self._extra_outputs.append(otpt)

        if costs is not None:
            # No cost or maximum flow on the additional links
            self.costs = costs + [0.0] * extra_slots
        if max_flows is not None:
            # The max_flows could be problematic with the aggregated node.
            self.max_flows = max_flows + [None] * extra_slots

        # Now create an aggregated node for addition constraints if required.
        if factors is not None:
            if extra_slots+1 != len(factors):
                raise ValueError("factors must have a length equal to extra_slots.")

            nodes = []
            valid_factors = []
            for r, nd in zip(factors, self.sublinks[n-1:]):
                if r is not None:
                    nodes.append(nd)
                    valid_factors.append(r)

            agg = AggregatedNode(self.model, "{} Agg".format(self.name), nodes)
            agg.factors = valid_factors

    def iter_slots(self, slot_name=None, is_connector=True):
        if is_connector:
            i = self.slot_names.index(slot_name)
            if i == 0:
                yield self.input
            else:
                yield self._extra_inputs[i-1]
        else:
            yield self.output


class AggregatedStorage(Loadable, Drawable, _core.AggregatedStorage, metaclass=NodeMeta):
    """ An aggregated sum of other `Storage` nodes

    This object should behave like `Storage` by returning current `flow`, `volume` and `current_pc`.
    However this object can not be connected to others within the network.

    Parameters
    ----------
    model - `Model` instance
    name - str
    storage_nodes - list or iterable of `Storage` objects
        The `Storage` objects which to return the sum total of

    Notes
    -----
    This node can not be connected to other nodes in the network.

    """
    __node_attributes__ = ('storage_nodes', )

    def __init__(self, model, name, storage_nodes, **kwargs):
        super(AggregatedStorage, self).__init__(model, name, **kwargs)
        self.storage_nodes = storage_nodes


class AggregatedNode(Loadable, Drawable, _core.AggregatedNode, metaclass=NodeMeta):
    """ An aggregated sum of other `Node` nodes

    This object should behave like `Node` by returning current `flow`.
    However this object can not be connected to others within the network.

    Parameters
    ----------
    model - `Model` instance
    name - str
    nodes - list or iterable of `Node` objects
        The `Node` objects which to return the sum total of

    Notes
    -----
    This node can not be connected to other nodes in the network.

    """
    __parameter_attributes__ = ('factors', 'min_flow', 'max_flow')
    __node_attributes__ = ('nodes', )

    def __init__(self, model, name, nodes, flow_weights=None, **kwargs):
        super(AggregatedNode, self).__init__(model, name, **kwargs)
        self.nodes = nodes
        self.flow_weights = flow_weights


class BreakLink(Node):
    """Compound node used to reduce the number of routes in a model

    Parameters
    ----------
    model : `pywr.model.Model`
    name : string
    min_flow : float or `pywr.parameters.Parameter`
    max_flow : float or `pywr.parameters.Parameter`
    cost : float or `pywr.parameters.Parameter`

    Notes
    -----

    In a model with form (3, 1, 3), i.e. 3 (A,B,C) inputs connected to 3
    outputs (D,E,F) via a bottleneck (X), there are 3*3 routes = 9 routes.

    ::

        A -->\\ /--> D
        B --> X --> E
        C -->/ \\--> F

    If X is a storage, there are only 6 routes: A->X_o, B->X_o, C->X_o and
    X_i->D_o, X_i->E_o, X_i->F_o.

    The `BreakLink` node is a compound node composed of a `Storage` with zero
    volume and a `Link`. It can be used in place of a normal `Link`, but
    with the benefit that it reduces the number of routes in the model (in
    the situation described above). The resulting LP is easier to solve.
    """
    allow_isolated = True

    def __init__(self, model, name, **kwargs):
        storage_name = "{} (storage)".format(name)
        link_name = "{} (link)".format(name)
        assert(storage_name not in model.nodes)
        assert(link_name not in model.nodes)
        self.storage = Storage(
            model,
            name=storage_name,
            min_volume=0,
            max_volume=0,
            initial_volume=0,
            cost=0,
        )
        self.link = Link(
            model,
            name=link_name
        )

        self.storage.connect(self.link)

        super(BreakLink, self).__init__(model, name, **kwargs)

    def min_flow():
        def fget(self):
            return self.link.min_flow

        def fset(self, value):
            self.link.min_flow = value
        return locals()
    min_flow = property(**min_flow())

    def max_flow():
        def fget(self):
            return self.link.max_flow

        def fset(self, value):
            self.link.max_flow = value
        return locals()
    max_flow = property(**max_flow())

    def cost():
        def fget(self):
            return self.link.cost

        def fset(self, value):
            self.link.cost = value
        return locals()
    cost = property(**cost())

    def iter_slots(self, slot_name=None, is_connector=True):
        if is_connector:
            # connecting FROM the transfer TO something else
            yield self.link
        else:
            # connecting FROM something else TO the transfer
            yield self.storage.outputs[0]

    def after(self, timestep):
        super(BreakLink, self).after(timestep)
        # update flow on transfer node to flow via link node
        self.commit_all(self.link.flow)


class DelayNode(Node):
    """ A node that delays flow for a given number of timesteps or days.

    This is a composite node consisting internally of an Input and an Output node. A
    `FlowDelayParameter` is used to delay the flow of the output node for a given period prior
    to this delayed flow being set as the flow of the input node. Connections to the node are connected
    to the internal output node and connection from the node are connected to the internal input node
    node.

    Parameters
    ----------
    model : `pywr.model.Model`
    name : string
        Name of the node.
    timesteps: int
        Number of timesteps to delay the flow.
    days: int
        Number of days to delay the flow. Specifying a number of days (instead of a number
        of timesteps) is only valid if the number of days is exactly divisible by the model
        timestep delta.
    initial_flow: float
        Flow provided by node for initial timesteps prior to any delayed flow being available.
        This is constant across all delayed timesteps and any model scenarios. Default is 0.0.
    """

    def __init__(self, model, name, **kwargs):
        self.allow_isolated = True
        output_name = "{} Output".format(name)
        input_name = "{} Input".format(name)
        param_name = "{} - delay parameter".format(name)
        assert(output_name not in model.nodes)
        assert(input_name not in model.nodes)
        assert(param_name not in model.parameters)

        days = kwargs.pop('days', 0)
        timesteps = kwargs.pop('timesteps', 0)
        initial_flow = kwargs.pop('initial_flow', 0.0)

        self.output = Output(model, name=output_name, parent=self)
        self.delay_param = FlowDelayParameter(model, self.output, timesteps=timesteps, days=days,
                                              initial_flow=initial_flow, name=param_name)
        self.input = Input(model, name=input_name, min_flow=self.delay_param, max_flow=self.delay_param, parent=self)
        super().__init__(model, name, **kwargs)

    def iter_slots(self, slot_name=None, is_connector=True):
        if is_connector:
            yield self.input
        else:
            yield self.output

    def after(self, timestep):
        super().after(timestep)
        # delayed flow is saved to the DelayNode
        self.commit_all(self.input.flow)


class LossLink(Node):
    """A node that has a proportional loss.

    A fixed proportional loss of all flow through this node is sent to an internal output node. Max and min flows
    applied to this node are enforced on the net output after losses. The node itself records the net output
    in its flow attribute (which would be used by any attached recorders).

    Parameters
    ----------
    model : `pywr.model.Model`
    name : string
        Name of the node.
    loss_factor : float
        The proportion of flow that is lost through this node. Must be greater than or equal to zero. If zero
        then no-losses are calculated. The percentage is calculated as a percentage of gross flow.
    """
    __parameter_value_attributes__ = ('loss_factor', )

    def __init__(self, model, name, **kwargs):
        self.allow_isolated = True

        output_name = "{} Output".format(name)
        gross_name = "{} Gross".format(name)
        net_name = "{} Net".format(name)
        agg_name = "{} Aggregated".format(name)

        assert(output_name not in model.nodes)
        assert(gross_name not in model.nodes)
        assert(net_name not in model.nodes)
        assert(agg_name not in model.nodes)

        self.output = Output(model, name=output_name, parent=self)
        self.gross = Link(model, name=gross_name, parent=self)
        self.net = Link(model, name=net_name, parent=self)
        self.gross.connect(self.output)
        self.gross.connect(self.net)

        self.agg = AggregatedNode(model, name=agg_name, nodes=[self.net, self.output])
        self.loss_factor = kwargs.pop('loss_factor', 0.0)

        super().__init__(model, name, **kwargs)

    def loss_factor():
        def fget(self):
            if self.agg.factors:
                return self.agg.factors[1]
            elif self.output.max_flow == 0.0:
                return 0.0
            else:
                return 1.0

        def fset(self, value):
            if value == 0.0:
                # 0% loss; no flow to the output loss node.
                self.agg.factors = None
                self.output.max_flow = 0.0
            elif value == 1.0:
                # 100% loss; all flow to the output loss node
                self.agg.factors = None
                self.output.max_flow = float('inf')
                self.net.max_flow = 0.0
            else:
                self.output.max_flow = float('inf')
                self.agg.factors = [1.0, float(value)]
        return locals()
    loss_factor = property(**loss_factor())

    def min_flow():
        def fget(self):
            return self.net.min_flow

        def fset(self, value):
            self.net.min_flow = value
        return locals()
    min_flow = property(**min_flow())

    def max_flow():
        def fget(self):
            return self.net.max_flow

        def fset(self, value):
            self.net.max_flow = value
        return locals()
    max_flow = property(**max_flow())

    def cost():
        def fget(self):
            return self.net.cost

        def fset(self, value):
            self.net.cost = value
        return locals()
    cost = property(**cost())

    def iter_slots(self, slot_name=None, is_connector=True):
        if is_connector:
            yield self.net
        else:
            yield self.gross

    def after(self, timestep):
        super().after(timestep)
        # Net flow is saved to the node.
        self.commit_all(self.net.flow)

from pywr.domains.river import *  # noqa



from pywr.nodes import Node

from fixtures import simple_linear_model, simple_storage_model

model = simple_linear_model


def test_keys_and_values(model):
    expected_names = {"Input", "Link", "Output"}
    assert set(model.nodes.keys()) == expected_names
    nodes = {node for node in model.nodes}
    assert len(nodes) == 3
    assert set(model.nodes.values()) == nodes


def test_contains(model):
    assert "Input" in model.nodes
    assert "Output" in model.nodes
    output = model.nodes["Output"]
    assert isinstance(output, Node)
    assert output in model.nodes


def test_delete(model):
    del(model.nodes["Output"])
    assert len(model.nodes) == 2
    assert {node.name for node in model.nodes} == {"Input", "Link"}


def test_delete_compound_node(simple_storage_model):
    """Removal of a compound node removes child nodes also"""
    assert len(list(simple_storage_model.nodes._nodes(hide_children=False))) == 5
    del(simple_storage_model.nodes["Storage"])
    assert len(list(simple_storage_model.nodes._nodes(hide_children=False))) == 2
    assert {node.name for node in simple_storage_model.nodes} == {"Input", "Output"}



#!/bin/python
"""
This script defines and runs a very simple model.

The network looks like this:

  A -->-- B -->-- C

"""
from pywr.model import Model
from pywr.nodes import Input, Output, Link

def create_model():
    # create a model
    model = Model(start="2016-01-01", end="2019-12-31", timestep=7)
    
    # create three nodes (an input, a link, and an output)
    A = Input(model, name="A", max_flow=10.0)
    B = Link(model, name="B", cost=10.0)
    C = Output(model, name="C", max_flow=5.0, cost=-20.0)
    
    # connect the nodes together
    A.connect(B)
    B.connect(C)
    
    return model

if __name__ == "__main__":
    # create the model and check it is valid
    model = create_model()
    model.check()
    
    # run the model
    result = model.run()
    print(result)
    
    # check the result of the model is as expected
    A = model.nodes["A"]
    assert(abs(A.flow[0] - 5.0) < 0.000001)
